"""Add extra quantities to Snap."""

from ..analysis import particles


def extra_quantities(snap):
    """Make extra quantities available.

    Parameters
    ----------
    snap
        The Snap object to add extra quantities to.
    """
    for array in particles.array_requires:
        _add_array(snap, array)


def _can_add_array(snap, name):
    return set(particles.array_requires[name]).issubset(snap.available_arrays())


def _vector(name):
    return True if name in particles.vector_arrays else False


def _dust(name):
    return True if name in particles.dust_arrays else False


def _add_array(snap, name):
    if _can_add_array(snap, name):
        vector = _vector(name)
        dust = _dust(name)
        snap._array_registry[name] = getattr(particles, name)
        if vector is True:
            snap._vector_arrays.add(name)
        if dust is True:
            snap._dust_arrays.add(name)



"""Phantom HDF5 snapshot file."""

from __future__ import annotations

from pathlib import Path
from typing import Callable, Dict, List, Union

import h5py
import numpy as np

from ..._config import load_config
from ..._logging import logger
from ..._units import Quantity, array_units, generate_array_code_units
from ..._units import units as plonk_units
from ..extra import extra_quantities
from ..snap import Snap
from ..utils import add_aliases

igas, iboundary, istar, idarkmatter, ibulge = 1, 3, 4, 5, 6
bignumber = 1e29


def fileid(snap: Snap) -> str:
    """Return Phantom 'fileid'.

    This contains information on the type of snapshot and which physics
    the simulation contains.

    Parameters
    ----------
    snap
        The Snap object.

    Returns
    -------
    str
        The file id.
    """
    return snap._file_pointer['header/fileident'][()].decode()


def generate_snap_from_file(
    filename: Union[str, Path], config: Union[str, Path] = None
) -> Snap:
    """Generate a Snap object from a Phantom HDF5 file.

    Parameters
    ----------
    filename
        The path to the file.
    config : optional
        The path to a Plonk config.toml file.

    Returns
    -------
    Snap
        A Snap object.
    """
    logger.debug(f'Loading Phantom snapshot: {filename}')

    # REQUIRED: Instantiate Snap.
    snap = Snap()

    # REQUIRED: Set snap.data_source.
    snap.data_source = 'Phantom'

    # REQUIRED: Set snap.file_path.
    file_path = Path(filename).expanduser()
    if not file_path.is_file():
        raise FileNotFoundError('Cannot find snapshot file')
    snap.file_path = file_path

    # REQUIRED: Set snap._file_pointer.
    file_handle = h5py.File(file_path, mode='r')
    snap._file_pointer = file_handle

    # REQUIRED: Set snap._properties, snap._code_units, and snap._array_code_units.
    header = {key: val[()] for key, val in file_handle['header'].items()}
    snap._properties, snap._code_units = header_to_properties(header)
    snap._array_code_units = generate_array_code_units(snap._code_units)

    # OPTIONAL: Set snap._units.
    snap._default_units = array_units(config=config)

    # OPTIONAL: Set snap._name_map.
    conf = load_config(filename=config)
    snap._name_map = {
        'particles': conf['phantom']['particles']['namemap'],
        'sinks': conf['phantom']['sinks']['namemap'],
    }

    # REQUIRED: Set snap._array_registry dictionary.
    #
    # The keys are the names of the arrays and the values are functions that return the
    # array when called with snap as the argument.
    arrays = list(file_handle['particles'])
    ndustsmall = header['ndustsmall']
    ndustlarge = header['ndustlarge']
    array_registry = populate_particle_array_registry(
        arrays=arrays,
        name_map=snap._name_map['particles'],
        ndustsmall=ndustsmall,
        ndustlarge=ndustlarge,
    )
    snap._array_registry.update(array_registry)

    # REQUIRED (if there are sink particles): Set _sink_registry.
    #
    # The keys are the names of the arrays and the values are functions that return the
    # array when called with snap as the argument.
    if header['nptmass'] > 0:
        sinks = list(file_handle['sinks'])
        sink_registry = populate_sink_array_registry(
            sinks=sinks, name_map=snap._name_map['sinks']
        )
        snap._sink_registry.update(sink_registry)

    # OPTIONAL: Add aliases
    add_aliases(snap)

    # OPTIONAL: Make extra derived quantities available.
    extra_quantities(snap)

    # REQUIRED: Return Snap object.
    return snap


def header_to_properties(header: dict):
    """Convert Phantom header to properties and units.

    Parameters
    ----------
    header
        The Phantom header as a dict.

    Returns
    -------
    prop
        The properties as a dict.
    units
        The units as a dict.
    """
    length = (header['udist'] * plonk_units('cm')).to_base_units()
    time = (header['utime'] * plonk_units('s')).to_base_units()
    mass = (header['umass'] * plonk_units('g')).to_base_units()
    temperature = plonk_units('K')
    magnetic_field = (
        header['umagfd']
        * plonk_units('g ** (1/2) / cm ** (1/2) / s')
        * np.sqrt(plonk_units.magnetic_constant / (4 * np.pi))
    ).to_base_units()
    current = (mass / time ** 2 / magnetic_field).to_base_units()

    units = {
        'length': length,
        'time': time,
        'mass': mass,
        'temperature': temperature,
        'current': current,
    }

    prop = dict()

    prop['time'] = header['time'] * units['time']
    prop['smoothing_length_factor'] = header['hfact']

    gamma = header['gamma']
    ieos = header['ieos']
    if ieos == 1:
        prop['equation_of_state'] = 'isothermal'
        prop['adiabatic_index'] = gamma
    elif ieos == 2:
        prop['equation_of_state'] = 'adiabatic'
        prop['adiabatic_index'] = gamma
    elif ieos == 3:
        prop['equation_of_state'] = 'locally isothermal disc'
        prop['adiabatic_index'] = gamma

    ndustsmall = header['ndustsmall']
    ndustlarge = header['ndustlarge']
    if ndustsmall > 0 and ndustlarge > 0:
        raise ValueError(
            'Phantom only supports either dust/gas mixtures (aka 1-fluid dust)\n'
            'or dust as separate sets of particles (aka multi-fluid dust).'
        )
    if ndustsmall > 0:
        prop['dust_method'] = 'dust/gas mixture'
    elif ndustlarge > 0:
        prop['dust_method'] = 'dust as separate sets of particles'

    n_dust = ndustsmall + ndustlarge
    if n_dust > 0:
        prop['grain_size'] = header['grainsize'][:n_dust] * units['length']
        prop['grain_density'] = (
            header['graindens'][:n_dust] * units['mass'] / units['length'] ** 3
        )

    return prop, units


def populate_particle_array_registry(
    arrays: List[str],
    name_map: Dict[str, str],
    ndustsmall: int = 0,
    ndustlarge: int = 0,
):
    """Populate particle arrays registry.

    Parameters
    ----------
    arrays
        A list of particle array names as they are on file.
    name_map
        A dict to convert from Phantom array names to Plonk names.
    ndustsmall
        The number of "small" dust species. I.e. the number of mixture
        method (1-fluid) dust species.
    ndustlarge
        The number of "large" dust species. I.e. the number of separate
        sets of particles (2-fluid) dust species.

    Returns
    -------
    Dict
        The particle array registry.
    """
    array_registry = dict()

    # Each particle gets an id
    array_registry['id'] = particle_id

    # Always read itype, xyz, h
    array_registry['type'] = particle_type
    array_registry['sub_type'] = sub_type
    array_registry['position'] = get_dataset('xyz', 'particles')
    array_registry['smoothing_length'] = get_dataset('h', 'particles')
    arrays.remove('itype')
    arrays.remove('xyz')
    arrays.remove('h')

    # Handle dust
    if ndustsmall > 0:
        array_registry['dust_fraction'] = dust_fraction
        arrays.remove('dustfrac')
    elif ndustlarge > 0:
        array_registry['dust_to_gas_ratio'] = dust_to_gas_ratio
        arrays.remove('dustfrac')
    if ndustsmall > 0 or ndustlarge > 0 and 'tstop' in arrays:
        array_registry['stopping_time'] = stopping_time
        arrays.remove('tstop')

    # Read arrays if available
    for name_on_file, name in name_map.items():
        if name_on_file in arrays:
            array_registry[name] = get_dataset(name_on_file, 'particles')
            arrays.remove(name_on_file)

    # Derived arrays not stored on file
    array_registry['mass'] = mass
    array_registry['density'] = density
    array_registry['pressure'] = pressure
    array_registry['sound_speed'] = sound_speed

    # Read *any* extra arrays
    for array in arrays:
        array_registry[array] = get_dataset(array, 'particles')

    return array_registry


def populate_sink_array_registry(sinks: List[str], name_map: Dict[str, str]):
    """Populate particle arrays registry.

    Parameters
    ----------
    sinks
        A list of sink array names as they are on file.
    name_map
        A dict to convert from Phantom array names to Plonk names.

    Returns
    -------
    Dict
        The sink array registry.
    """
    sink_registry = dict()

    # Read arrays if available
    for name_on_file, name in name_map.items():
        if name_on_file in sinks:
            sink_registry[name] = get_dataset(name_on_file, 'sinks')
            sinks.remove(name_on_file)

    # Read *any* extra arrays
    for sink in sinks:
        sink_registry[sink] = get_dataset(sink, 'sinks')

    return sink_registry


def get_dataset(dataset: str, group: str) -> Callable:
    """Return a function that returns an array from file.

    Parameters
    ----------
    dataset
        The name of the HDF5 dataset, i.e. the array name.
    group
        The name of the HDF5 group. For Phantom, this is one of
        'header', 'particles', or 'sinks'.

    Returns
    -------
    Callable
        The function that when called with a Snap object returns the
        array in the dataset with appropriate units, i.e. as a Pint
        Quantity.
    """

    def func(snap: Snap) -> Quantity:
        array = snap._file_pointer[f'{group}/{dataset}'][()]
        name_map = snap._name_map[group]
        if dataset in name_map:
            name = name_map[dataset]
        else:
            name = dataset
        try:
            unit = snap._array_code_units[name]
        except KeyError:
            logger.error(
                f'Cannot get unit of dataset "{group}/{dataset}" - '
                'assuming dimensionless'
            )
            unit = plonk_units('dimensionless')
        return array * unit

    return func


def particle_id(snap: Snap) -> Quantity:
    """Particle id."""
    num_particles = snap._file_pointer['header/nparttot'][()]
    return np.arange(num_particles) * plonk_units('dimensionless')


def particle_type(snap: Snap) -> Quantity:
    """Particle type.

    Type        | Phantom                               | Plonk
    ----------- | ------------------------------------- | -----
    Gas         |                                     1 | 1
    Dust        |      idust -> idust + ndustlarge      | 2
    Boundary    | idustbound -> idustbound + ndustlarge | 3
    Star        |                                     4 | 4
    Dark matter |                                     5 | 5
    Bulge       |                                     6 | 6
    """
    idust = snap._file_pointer['header/idust'][()]
    ndustlarge = snap._file_pointer['header/ndustlarge'][()]
    particle_type = np.abs(get_dataset('itype', 'particles')(snap))
    particle_type[
        (particle_type >= idust) & (particle_type < idust + ndustlarge)
    ] = snap.particle_type['dust']
    try:
        idustbound = snap._file_pointer['header/idustbound'][()]
        particle_type[
            (particle_type >= idustbound) & (particle_type < idustbound + ndustlarge)
        ] = snap.particle_type['boundary']
    except KeyError:
        if np.any(particle_type >= idust + ndustlarge):
            logger.error('Cannot determine dust boundary particles')
    return np.array(particle_type.magnitude, dtype=int) * plonk_units('dimensionless')


def sub_type(snap: Snap) -> Quantity:
    """Particle sub-type.

              | Types
              | Gas | Dust | Boundary
    Sub-types |     |      |
    --------- | --- | ---- | --------
    Gas       |  0  |  n/a |  0
    Dust 1    | n/a |   1  |  1
    Dust 2    | n/a |   2  |  2
    Dust 3    | n/a |   3  |  3
    ...
    """
    particle_type = np.abs(get_dataset('itype', 'particles')(snap))
    sub_type = np.zeros(particle_type.shape, dtype=np.int8)
    sub_type[
        (particle_type == igas)
        | (particle_type == istar)
        | (particle_type == idarkmatter)
        | (particle_type == ibulge)
    ] = 0
    sub_type[particle_type == iboundary] = 0
    idust = snap._file_pointer['header/idust'][()]
    ndustlarge = snap._file_pointer['header/ndustlarge'][()]
    for idx in range(idust, idust + ndustlarge):
        sub_type[particle_type == idx] = idx - idust + 1
    try:
        idustbound = snap._file_pointer['header/idustbound'][()]
        for idx in range(idustbound, idustbound + ndustlarge):
            sub_type[particle_type == idx] = idx - idustbound + 1
    except KeyError:
        if np.any(particle_type >= idust + ndustlarge):
            logger.error('Cannot determine dust boundary particles')
    return sub_type * plonk_units('dimensionless')


def mass(snap: Snap) -> Quantity:
    """Particle mass."""
    massoftype = snap._file_pointer['header/massoftype'][()]
    particle_type = np.array(
        np.abs(get_dataset('itype', 'particles')(snap)).magnitude, dtype=int
    )
    return massoftype[particle_type - 1] * snap._array_code_units['mass']


def density(snap: Snap) -> Quantity:
    """Density."""
    m = (mass(snap) / snap._array_code_units['mass']).magnitude
    h = (
        get_dataset('h', 'particles')(snap) / snap._array_code_units['smoothing_length']
    ).magnitude
    hfact = snap.properties['smoothing_length_factor']
    rho = m * (hfact / np.abs(h)) ** 3
    return rho * snap._array_code_units['density']


def pressure(snap: Snap) -> Quantity:
    """Pressure."""
    ieos = snap._file_pointer['header/ieos'][()]
    K = 2 / 3 * snap._file_pointer['header/RK2'][()]
    gamma = snap.properties['adiabatic_index']
    rho = density(snap)
    if ieos == 1:
        # Globally isothermal
        K = K * snap.code_units['length'] ** 2 * snap.code_units['time'] ** (-2)
        return K * rho
    if ieos == 2:
        # Adiabatic
        try:
            energy = get_dataset('u', 'particles')(snap)
            if gamma > 1.0001:
                return (gamma - 1) * energy * rho
            else:
                return 2 / 3 * energy * rho
        except KeyError:
            K = (
                K
                * snap.code_units['length'] ** (1 - gamma)
                * snap.code_units['mass'] ** (-1 + 3 * gamma)
                * snap.code_units['time'] ** (-2)
            )
            return K * rho ** (gamma - 1)
    if ieos == 3:
        # Vertically isothermal (for accretion disc)
        K = K * snap.code_units['length'] ** 2 * snap.code_units['time'] ** (-2)
        q = snap._file_pointer['header/qfacdisc'][()]
        pos = get_dataset('xyz', 'particles')(snap)
        r_squared = pos[:, 0] ** 2 + pos[:, 1] ** 2 + pos[:, 2] ** 2
        r_squared = (r_squared / snap._array_code_units['position'] ** 2).magnitude
        return K * rho * r_squared ** (-q)
    raise ValueError('Unknown equation of state')


def sound_speed(snap: Snap) -> Quantity:
    """Sound speed."""
    ieos = snap._file_pointer['header/ieos'][()]
    gamma = snap.properties['adiabatic_index']
    rho = density(snap)
    P = pressure(snap)
    if ieos in (1, 3):
        return np.sqrt(P / rho)
    if ieos == 2:
        return np.sqrt(gamma * P / rho)
    raise ValueError('Unknown equation of state')


def stopping_time(snap: Snap) -> Quantity:
    """Dust stopping time."""
    stopping_time = get_dataset('tstop', 'particles')(snap)
    stopping_time[stopping_time == bignumber] = np.inf * snap.code_units['time']
    return stopping_time


def dust_fraction(snap: Snap) -> Quantity:
    """Dust fraction for mixture method (1-fluid)."""
    if snap.properties['dust_method'] != 'dust/gas mixture':
        raise ValueError('Dust fraction only available for "dust/gas mixture"')
    return get_dataset('dustfrac', 'particles')(snap)


def dust_to_gas_ratio(snap: Snap) -> Quantity:
    """Dust-to-gas ratio for separate particles method (2-fluid)."""
    if snap.properties['dust_method'] != 'dust as separate sets of particles':
        raise ValueError(
            'Dust fraction only available for "dust as separate sets of particles"'
        )
    return get_dataset('dustfrac', 'particles')(snap)



"""Interpolation to a pixel grid.

There are two functions: one for interpolation of scalar fields, and one
for interpolation of vector fields.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Tuple

import numpy as np
from numpy import ndarray

from .._logging import logger
from ..utils.geometry import distance_from_plane
from .splash import interpolate_projection, interpolate_slice

if TYPE_CHECKING:
    from ..snap.snap import SnapLike

Extent = Tuple[float, float, float, float]


def interpolate(
    *,
    snap: SnapLike,
    quantity: str,
    x: str = 'x',
    y: str = 'y',
    interp: 'str',
    slice_normal: Tuple[float, float, float] = None,
    slice_offset: float = None,
    extent: Extent,
    **kwargs,
) -> ndarray:
    """Interpolate a quantity on the snapshot to a pixel grid.

    Parameters
    ----------
    snap
        The Snap (or SubSnap) object.
    quantity
        The quantity to visualize. Must be a string to pass to Snap,
    x
        The x-coordinate for the visualization. Must be a string to
        pass to Snap. Default is 'x'.
    y
        The y-coordinate for the visualization. Must be a string to
        pass to Snap. Default is 'y'.
    interp
        The interpolation type. Default is 'projection'.

        - 'projection' : 2d interpolation via projection to xy-plane
        - 'slice' : 3d interpolation via cross-section slice.
    slice_normal
        The normal vector to the plane in which to take the
        cross-section slice as an array (x, y, z).
    slice_offset
        The offset of the cross-section slice. Default is 0.0.
    extent
        The xy extent of the image as (xmin, xmax, ymin, ymax).
    **kwargs
        Additional keyword arguments to pass to scalar_interpolation
        and vector_interpolation.

    Returns
    -------
    ndarray
        The interpolated quantity on a pixel grid as an ndarray. The
        shape for scalar data is (npixx, npixy), and for vector is
        (2, npixx, npixy).

    Examples
    --------
    Interpolate density to grid.

    >>> grid_data = plonk.interpolate(
    ...     snap=snap,
    ...     quantity='density',
    ...     interp='projection',
    ...     extent=(-100, 100, -100, 100),
    ... )
    """
    _quantity, x, y, z = _get_arrays_from_str(snap=snap, quantity=quantity, x=x, y=y)
    h = snap.array_in_code_units('smoothing_length')
    m = snap.array_in_code_units('mass')

    if interp == 'projection':
        dist_from_slice = None
        if slice_normal is not None:
            logger.warning('ignoring slice_normal for projection')
        if slice_offset is not None:
            logger.warning('ignoring slice_offset for projection')
    elif interp == 'slice':
        if slice_offset is None:
            slice_offset = 0.0
        if slice_normal is None:
            slice_normal = np.array([0, 0, 1])
        dist_from_slice = distance_from_plane(x, y, z, slice_normal, slice_offset)

    if _quantity.ndim == 1:
        interpolated_data = scalar_interpolation(
            quantity=_quantity,
            x_coordinate=x,
            y_coordinate=y,
            dist_from_slice=dist_from_slice,
            extent=extent,
            smoothing_length=h,
            particle_mass=m,
            hfact=snap.properties['smoothing_length_factor'],
            **kwargs,
        )

    elif _quantity.ndim == 2:
        interpolated_data = vector_interpolation(
            quantity_x=_quantity[:, 0],
            quantity_y=_quantity[:, 1],
            x_coordinate=x,
            y_coordinate=y,
            dist_from_slice=dist_from_slice,
            extent=extent,
            smoothing_length=h,
            particle_mass=m,
            hfact=snap.properties['smoothing_length_factor'],
            **kwargs,
        )

    else:
        raise ValueError('quantity.ndim > 2: cannot determine quantity')

    return interpolated_data


def scalar_interpolation(
    *,
    quantity: ndarray,
    x_coordinate: ndarray,
    y_coordinate: ndarray,
    dist_from_slice: ndarray = None,
    extent: Extent,
    smoothing_length: ndarray,
    particle_mass: ndarray,
    hfact: float,
    number_of_pixels: Tuple[float, float] = (512, 512),
    density_weighted: bool = None,
) -> ndarray:
    """Interpolate scalar quantity to a pixel grid.

    Parameters
    ----------
    quantity
        A scalar quantity on the particles to interpolate.
    x_coordinate
        Particle coordinate for x-axis in interpolation.
    y_coordinate
        Particle coordinate for y-axis in interpolation.
    dist_from_slice
        The distance from the cross section slice. Only required for
        cross section interpolation.
    extent
        The range in the x- and y-direction as (xmin, xmax, ymin, ymax).
    smoothing_length
        The smoothing length on each particle.
    particle_mass
        The particle mass on each particle.
    hfact
        The smoothing length factor.
    number_of_pixels
        The pixel grid to interpolate the scalar quantity to, as
        (npixx, npixy).
    density_weighted
        Use density weighted interpolation. Default is off.

    Returns
    -------
    ndarray
        An array of scalar quantities interpolated to a pixel grid with
        shape (npixx, npixy).
    """
    return _interpolate(
        quantity=quantity,
        x_coordinate=x_coordinate,
        y_coordinate=y_coordinate,
        dist_from_slice=dist_from_slice,
        extent=extent,
        smoothing_length=smoothing_length,
        particle_mass=particle_mass,
        hfact=hfact,
        number_of_pixels=number_of_pixels,
        density_weighted=density_weighted,
    )


def vector_interpolation(
    *,
    quantity_x: ndarray,
    quantity_y: ndarray,
    x_coordinate: ndarray,
    y_coordinate: ndarray,
    dist_from_slice: ndarray = None,
    extent: Extent,
    smoothing_length: ndarray,
    particle_mass: ndarray,
    hfact: float,
    number_of_pixels: Tuple[float, float] = (512, 512),
    density_weighted: bool = None,
) -> ndarray:
    """Interpolate scalar quantity to a pixel grid.

    Parameters
    ----------
    quantity_x
        The x-component of a vector quantity to interpolate.
    quantity_y
        The y-component of a vector quantity to interpolate.
    x_coordinate
        Particle coordinate for x-axis in interpolation.
    y_coordinate
        Particle coordinate for y-axis in interpolation.
    dist_from_slice
        The distance from the cross section slice. Only required for
        cross section interpolation.
    extent
        The range in the x- and y-direction as (xmin, xmax, ymin, ymax).
    smoothing_length
        The smoothing length on each particle.
    particle_mass
        The particle mass on each particle.
    hfact
        The smoothing length factor.
    number_of_pixels
        The pixel grid to interpolate the scalar quantity to, as
        (npixx, npixy).
    density_weighted
        Use density weighted interpolation. Default is off.

    Returns
    -------
    ndarray
        An array of vector quantities interpolated to a pixel grid with
        shape (2, npixx, npixy).
    """
    vecsmoothx = _interpolate(
        quantity=quantity_x,
        x_coordinate=x_coordinate,
        y_coordinate=y_coordinate,
        dist_from_slice=dist_from_slice,
        extent=extent,
        smoothing_length=smoothing_length,
        particle_mass=particle_mass,
        hfact=hfact,
        number_of_pixels=number_of_pixels,
        density_weighted=density_weighted,
    )
    vecsmoothy = _interpolate(
        quantity=quantity_y,
        x_coordinate=x_coordinate,
        y_coordinate=y_coordinate,
        dist_from_slice=dist_from_slice,
        extent=extent,
        smoothing_length=smoothing_length,
        particle_mass=particle_mass,
        hfact=hfact,
        number_of_pixels=number_of_pixels,
        density_weighted=density_weighted,
    )
    return np.stack((np.array(vecsmoothx), np.array(vecsmoothy)))


def _interpolate(
    *,
    quantity: ndarray,
    x_coordinate: ndarray,
    y_coordinate: ndarray,
    dist_from_slice: ndarray = None,
    extent: Extent,
    smoothing_length: ndarray,
    particle_mass: ndarray,
    hfact: float,
    number_of_pixels: Tuple[float, float],
    density_weighted: bool = None,
) -> ndarray:
    if dist_from_slice is None:
        do_slice = False
    else:
        do_slice = True
    normalise = False
    if density_weighted is None:
        density_weighted = False
    if density_weighted:
        normalise = True

    npixx, npixy = number_of_pixels
    xmin, ymin = extent[0], extent[2]
    pixwidthx = (extent[1] - extent[0]) / npixx
    pixwidthy = (extent[3] - extent[2]) / npixy
    npart = len(smoothing_length)

    itype = np.ones(smoothing_length.shape)
    if density_weighted:
        weight = particle_mass / smoothing_length ** 3
    else:
        weight = hfact ** -3 * np.ones(smoothing_length.shape)

    if do_slice:
        interpolated_data = interpolate_slice(
            x=x_coordinate,
            y=y_coordinate,
            dslice=dist_from_slice,
            hh=smoothing_length,
            weight=weight,
            dat=quantity,
            itype=itype,
            npart=npart,
            xmin=xmin,
            ymin=ymin,
            npixx=npixx,
            npixy=npixy,
            pixwidthx=pixwidthx,
            pixwidthy=pixwidthy,
            normalise=normalise,
        )
    else:
        interpolated_data = interpolate_projection(
            x=x_coordinate,
            y=y_coordinate,
            hh=smoothing_length,
            weight=weight,
            dat=quantity,
            itype=itype,
            npart=npart,
            xmin=xmin,
            ymin=ymin,
            npixx=npixx,
            npixy=npixy,
            pixwidthx=pixwidthx,
            pixwidthy=pixwidthy,
            normalise=normalise,
        )

    return interpolated_data


def _get_arrays_from_str(*, snap, quantity, x, y):

    coords = {'x', 'y', 'z'}
    if x not in coords:
        raise ValueError('x-coordinate must be one of "x", "y", "z"')
    if y not in coords:
        raise ValueError('y-coordinate must be one of "x", "y", "z"')

    quantity_str, x_str, y_str = quantity, x, y
    z_str = coords.difference((x_str, y_str)).pop()

    quantity = snap.array_in_code_units(quantity_str)
    x = snap.array_in_code_units(x_str)
    y = snap.array_in_code_units(y_str)
    z = snap.array_in_code_units(z_str)

    if quantity.ndim > 2:
        raise ValueError('Cannot interpret quantity with ndim > 2')
    if quantity.ndim == 2:
        try:
            quantity_x = snap.array_in_code_units(quantity_str + '_' + x_str)
            quantity_y = snap.array_in_code_units(quantity_str + '_' + y_str)
            quantity = np.stack([quantity_x, quantity_y]).T
        except ValueError:
            raise ValueError(
                '2d quantity must be a vector quantity, e.g. "velocity".\n'
                'For dust quantities, try appending the dust species number,\n'
                'e.g. "dust_density_001".'
            )

    return quantity, x, y, z



"""Extra pre-defined profiles."""

from functools import partial

import numpy as np

from .._units import Quantity
from .._units import units as plonk_units

G = (1 * plonk_units.newtonian_constant_of_gravitation).to_base_units()


def extra_profiles(profile, num_mixture_dust_species: int = 0):
    """Make extra profiles available.

    Parameters
    ----------
    profile
        The profile object to add extra profiles to.
    num_mixture_dust_species
        The number of "mixture" dust species.
    """

    @profile.add_profile
    def mass(prof) -> Quantity:
        """Mass profile."""
        M = prof.snap['mass']
        return prof.particles_to_binned_quantity('sum', M)

    @profile.add_profile
    def surface_density(prof) -> Quantity:
        """Surface density profile.

        Units are [mass / length ** ndim], which depends on ndim of profile.
        """
        return prof['mass'] / prof['size']

    @profile.add_profile
    def scale_height(prof) -> Quantity:
        """Scale height profile."""
        z = prof.snap['z']
        return prof.particles_to_binned_quantity('std', z)

    @profile.add_profile
    def aspect_ratio(prof) -> Quantity:
        """Aspect ratio profile."""
        H = prof['scale_height']
        R = prof['radius']
        return H / R

    @profile.add_profile
    def angular_momentum_theta(prof) -> Quantity:
        """Angle between specific angular momentum and xy-plane."""
        angular_momentum_z = prof['angular_momentum_z']
        angular_momentum_magnitude = prof['angular_momentum_mag']

        return np.arccos(angular_momentum_z / angular_momentum_magnitude)

    @profile.add_profile
    def angular_momentum_phi(prof) -> Quantity:
        """Angle between specific angular momentum and x-axis in xy-plane."""
        angular_momentum_x = prof['angular_momentum_x']
        angular_momentum_y = prof['angular_momentum_y']
        return np.arctan2(angular_momentum_y, angular_momentum_x)

    @profile.add_profile
    def toomre_Q(prof) -> Quantity:
        """Toomre Q parameter."""
        return (
            prof['sound_speed']
            * prof['keplerian_frequency']
            / (np.pi * G * prof['surface_density'])
        )

    if num_mixture_dust_species > 0:

        @profile.add_profile
        def gas_mass(prof) -> Quantity:
            """Gas mass profile."""
            M = prof.snap['gas_mass']
            return prof.particles_to_binned_quantity('sum', M)

        @profile.add_profile
        def gas_surface_density(prof) -> Quantity:
            """Gas surface density profile.

            Units are [mass / length ** ndim], which depends on ndim of profile.
            """
            return prof['gas_mass'] / prof['size']

        for idx in range(num_mixture_dust_species):

            def dust_mass(idx, prof) -> Quantity:
                """Dust mass profile."""
                M = prof.snap[f'dust_mass_{idx+1:03}']
                return prof.particles_to_binned_quantity('sum', M)

            def dust_surface_density(idx, prof) -> Quantity:
                """Dust surface density profile.

                Units are [mass / length ** ndim], which depends on ndim of profile.
                """
                return prof[f'dust_mass_{idx+1:03}'] / prof['size']

            profile._profile_functions[f'dust_mass_{idx+1:03}'] = partial(
                dust_mass, idx
            )

            profile._profile_functions[f'dust_surface_density_{idx+1:03}'] = partial(
                dust_surface_density, idx
            )



"""Visualize smoothed particle hydrodynamics data."""

from __future__ import annotations

from copy import copy
from typing import TYPE_CHECKING, Any, Dict, Sequence, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
from mpl_toolkits.axes_grid1 import make_axes_locatable
from numpy import ndarray

from .._logging import logger
from .._units import Quantity
from .._units import units as plonk_units
from ..utils.utils import pretty_array_name
from . import plots
from .functions import get_extent_from_percentile
from .interpolation import interpolate

if TYPE_CHECKING:
    from ..snap.snap import SnapLike

_kind_to_function = {
    'image': plots.imshow,
    'contour': plots.contour,
    'quiver': plots.quiver,
    'streamplot': plots.streamplot,
}


def image(
    snap: SnapLike,
    quantity: str,
    *,
    x: str = 'x',
    y: str = 'y',
    interp: str = 'projection',
    slice_normal: Tuple[float, float, float] = None,
    slice_offset: Union[Quantity, float] = None,
    extent: Quantity = None,
    units: Dict[str, str] = None,
    ax: Any = None,
    ax_kwargs={},
    colorbar_kwargs={},
    **kwargs,
) -> Any:
    """Visualize scalar SPH data as an image.

    Visualize scalar smoothed particle hydrodynamics data by
    interpolation to a pixel grid.

    Parameters
    ----------
    snap
        The Snap (or SubSnap) object to visualize.
    quantity
        The quantity to visualize. Must be a string to pass to Snap.
    x
        The x-coordinate for the visualization. Must be a string to
        pass to Snap. Default is 'x'.
    y
        The y-coordinate for the visualization. Must be a string to
        pass to Snap. Default is 'y'.
    interp
        The interpolation type. Default is 'projection'.

        - 'projection' : 2d interpolation via projection to xy-plane
        - 'slice' : 3d interpolation via cross-section slice.
    slice_normal
        The normal vector to the plane in which to take the
        cross-section slice as a tuple (x, y, z). Default is
        (0, 0, 1).
    slice_offset
        The offset of the cross-section slice. Default is 0.0.
    extent
        The range in the x and y-coord as (xmin, xmax, ymin, ymax)
        where xmin, etc. can be floats or quantities with units of
        length. The default is to set the extent to a box of size such
        that 99% of particles are contained within.
    units
        The units of the plot as a dictionary. The keys correspond to
        quantities such as 'position', 'density', 'velocity', and so on.
        The values are strings representing units, e.g. 'g/cm^3' for
        density. There is a special key 'projection' that corresponds
        to the length unit in the direction of projection for projected
        interpolation plots.
    ax
        A matplotlib Axes handle.
    ax_kwargs
        Keyword arguments to pass to matplotlib Axes.
    colorbar_kwargs
        Keyword arguments to pass to matplotlib Colorbar.
    **kwargs
        Additional keyword arguments to pass to interpolation and
        matplotlib functions.

    Returns
    -------
    ax
        The matplotlib Axes object.

    Notes
    -----
    Additional parameters passed as keyword arguments will be
    passed to lower level functions as required. E.g. Plonk uses
    matplotlib's imshow for a image plot, so additional arguments
    to imshow can be passed this way.

    See below for additional parameters for interpolation,
    colorbars, etc. All other keyword arguments are passed to the
    appropriate matplotlib function.

    Other Parameters
    ----------------
    number_of_pixels : tuple
        The number of pixels to interpolate particle quantities
        to as a tuple (nx, ny). Default is (512, 512).
    density_weighted : bool
        Whether to density weight the interpolation or not.
        Default is False.
    show_colorbar : bool
        Whether or not to display a colorbar. Default is True.

    Examples
    --------
    Show an image of the surface density in xy-plane.

    >>> plonk.image(snap=snap, quantity='density')

    Alternatively, access the function as a method on the Snap object.

    >>> snap.image(quantity='density')

    Set units for the plot.

    >>> units = {'position': 'au', 'density': 'g/cm^3', 'projection': 'cm'}
    >>> snap.image(quantity='density', units=units)

    Show a slice image of the density in xy-plane at z=0.

    >>> snap.image(quantity='density', interp='slice')
    """
    return _interpolation_plot(
        snap=snap,
        quantity=quantity,
        x=x,
        y=y,
        kind='image',
        interp=interp,
        slice_normal=slice_normal,
        slice_offset=slice_offset,
        extent=extent,
        units=units,
        ax=ax,
        ax_kwargs=ax_kwargs,
        colorbar_kwargs=colorbar_kwargs,
        **kwargs,
    )


def vector(
    snap: SnapLike,
    quantity: str,
    *,
    x: str = 'x',
    y: str = 'y',
    interp: str = 'projection',
    slice_normal: Tuple[float, float, float] = None,
    slice_offset: Union[Quantity, float] = None,
    extent: Quantity = None,
    units: Dict[str, str] = None,
    ax: Any = None,
    ax_kwargs={},
    **kwargs,
) -> Any:
    """Visualize vector SPH data as a vector plot.

    Visualize scalar smoothed particle hydrodynamics data by
    interpolation to a pixel grid of arrows.

    Parameters
    ----------
    snap
        The Snap (or SubSnap) object to visualize.
    quantity
        The quantity to visualize. Must be a string to pass to Snap.
    x
        The x-coordinate for the visualization. Must be a string to
        pass to Snap. Default is 'x'.
    y
        The y-coordinate for the visualization. Must be a string to
        pass to Snap. Default is 'y'.
    interp
        The interpolation type. Default is 'projection'.

        - 'projection' : 2d interpolation via projection to xy-plane
        - 'slice' : 3d interpolation via cross-section slice.
    slice_normal
        The normal vector to the plane in which to take the
        cross-section slice as a tuple (x, y, z). Default is
        (0, 0, 1).
    slice_offset
        The offset of the cross-section slice. Default is 0.0.
    extent
        The range in the x and y-coord as (xmin, xmax, ymin, ymax)
        where xmin, etc. can be floats or quantities with units of
        length. The default is to set the extent to a box of size such
        that 99% of particles are contained within.
    units
        The units of the plot as a dictionary. The keys correspond to
        quantities such as 'position', 'density', 'velocity', and so on.
        The values are strings representing units, e.g. 'g/cm^3' for
        density. There is a special key 'projection' that corresponds
        to the length unit in the direction of projection for projected
        interpolation plots.
    ax
        A matplotlib Axes handle.
    ax_kwargs
        Keyword arguments to pass to matplotlib Axes.
    **kwargs
        Additional keyword arguments to pass to interpolation and
        matplotlib functions.

    Returns
    -------
    ax
        The matplotlib Axes object.

    Notes
    -----
    Additional parameters passed as keyword arguments will be
    passed to lower level functions as required.

    See below for additional parameters for interpolation, vector
    properties, etc. All other keyword arguments are passed to the
    appropriate matplotlib function.

    Other Parameters
    ----------------
    number_of_pixels : tuple
        The number of pixels to interpolate particle quantities
        to as a tuple (nx, ny). Default is (512, 512).
    density_weighted : bool
        Whether to density weight the interpolation or not.
        Default is False.
    number_of_arrows : tuple
        The number of arrows to display by sub-sampling the
        interpolated data. Default is (25, 25).
    normalize_vectors : bool
        Whether to normalize the arrows to all have the same
        length. Default is False.

    Examples
    --------
    Show a vector plot of velocity in xy-plane.

    >>> plonk.vector(snap=snap, quantity='velocity')

    Alternatively, access the function as a method on the Snap object.

    >>> snap.vector(quantity='velocity')

    Set units for the plot.

    >>> units = {'position': 'au', 'velocity': 'km/s', 'projection': 'km'}
    >>> snap.vector(quantity='velocity', units=units)

    Show a slice plot of the velocity in xy-plane at z=0.

    >>> snap.vector(quantity='density', interp='slice')
    """
    return _interpolation_plot(
        snap=snap,
        quantity=quantity,
        x=x,
        y=y,
        kind='quiver',
        interp=interp,
        slice_normal=slice_normal,
        slice_offset=slice_offset,
        extent=extent,
        units=units,
        ax=ax,
        ax_kwargs=ax_kwargs,
        **kwargs,
    )


def _interpolation_plot(
    snap: SnapLike,
    quantity: str,
    *,
    x: str = 'x',
    y: str = 'y',
    kind: str = None,
    interp: str = 'projection',
    slice_normal: Tuple[float, float, float] = None,
    slice_offset: Union[Quantity, float] = None,
    extent: Quantity = None,
    units: Dict[str, str] = None,
    ax: Any = None,
    ax_kwargs={},
    colorbar_kwargs={},
    **kwargs,
) -> Any:
    logger.debug(f'Visualizing "{quantity}" on snap: {snap.file_path.name}')
    _kwargs = copy(kwargs)

    if interp not in ('projection', 'slice'):
        raise ValueError('interp must be "projection" or "slice"')

    q: Quantity = snap[quantity]
    if kind is None:
        if q.ndim == 1:
            kind = 'image'
        elif q.ndim == 2:
            kind = 'quiver'
    if kind not in ('image', 'contour', 'quiver', 'streamplot'):
        raise ValueError('Cannot determine plot type')
    if kind in ('image', 'contour') and not q.ndim == 1:
        raise ValueError('image and contour plots must be of 1-dimensional quantities')
    if kind in ('quiver', 'streamplot') and quantity not in snap._vector_arrays:
        raise ValueError('quiver and stream plots must be of vector quantities')

    if ax is None:
        fig, ax = plt.subplots()
    else:
        fig = ax.figure

    # Interpolate data to plot
    interp_kwargs = {
        key: val
        for key, val in _kwargs.items()
        if key in ('number_of_pixels', 'density_weighted')
    }
    for key in interp_kwargs:
        _kwargs.pop(key)
    _data, _extent, _units = _interpolated_data(
        snap=snap,
        quantity=quantity,
        x=x,
        y=y,
        interp=interp,
        slice_normal=slice_normal,
        slice_offset=slice_offset,
        extent=extent,
        units=units,
        **interp_kwargs,
    )

    # Make the actual plot
    _interpolated_plot(
        interpolated_data=_data,
        extent=_extent,
        names={'quantity': quantity, 'x': x, 'y': y},
        kind=kind,
        interp=interp,
        units=_units,
        ax=ax,
        fig=fig,
        ax_kwargs=ax_kwargs,
        colorbar_kwargs=colorbar_kwargs,
        **_kwargs,
    )

    return ax


def _interpolated_data(
    snap, quantity, x, y, interp, slice_normal, slice_offset, extent, units, **kwargs,
):
    _units = {
        'quantity': _get_unit(snap, quantity, units),
        'extent': _get_unit(snap, 'position', units),
        'projection': _get_unit(snap, 'projection', units),
    }

    if extent is None:
        _extent = get_extent_from_percentile(snap=snap, x=x, y=y)
        _extent = np.array([e.magnitude for e in _extent]) * _extent[0].units
    else:
        _extent = extent
    if not isinstance(_extent[0], Quantity):
        _extent = (
            _extent[0] * _units['extent'],
            _extent[1] * _units['extent'],
            _extent[2] * _units['extent'],
            _extent[3] * _units['extent'],
        )
    __extent = (
        (_extent[0] / snap.code_units['length']).to_base_units().magnitude,
        (_extent[1] / snap.code_units['length']).to_base_units().magnitude,
        (_extent[2] / snap.code_units['length']).to_base_units().magnitude,
        (_extent[3] / snap.code_units['length']).to_base_units().magnitude,
    )

    if interp == 'slice':
        if slice_normal is None:
            slice_normal = (0, 0, 1)
        if slice_offset is None:
            slice_offset = 0 * plonk_units('cm')
        if isinstance(slice_offset, Quantity):
            slice_offset = (
                (slice_offset / snap.code_units['length']).to_base_units().magnitude
            )
        else:
            logger.warning('slice_offset has no units, assuming code units')

    # Interpolate in code units
    interpolated_data = interpolate(
        snap=snap,
        quantity=quantity,
        x=x,
        y=y,
        interp=interp,
        slice_normal=slice_normal,
        slice_offset=slice_offset,
        extent=__extent,
        **kwargs,
    )
    # Convert back to physical units
    interpolated_data, ___extent = _convert_units_for_interpolation(
        snap=snap,
        quantity=quantity,
        interpolated_data=interpolated_data,
        extent=__extent,
        units=_units,
        interp=interp,
    )

    return interpolated_data, ___extent, _units


def _interpolated_plot(
    interpolated_data,
    extent,
    names,
    kind,
    interp,
    units,
    ax,
    fig,
    ax_kwargs,
    colorbar_kwargs,
    **kwargs,
):

    show_colorbar = kwargs.pop('show_colorbar', kind == 'image')

    vmin, vmax = kwargs.get('vmin', None), kwargs.get('vmax', None)
    vmin = _convert_units_for_cmap(vmin, 'vmin', units, interp)
    vmax = _convert_units_for_cmap(vmax, 'vmax', units, interp)
    if vmin is not None:
        kwargs['vmin'] = vmin
    if vmax is not None:
        kwargs['vmax'] = vmax

    plot_object = _kind_to_function[kind](
        interpolated_data=interpolated_data, extent=extent, ax=ax, **kwargs,
    )

    ax.set_xlim(*extent[:2])
    ax.set_ylim(*extent[2:])

    eunit = units['extent']
    if np.allclose(eunit.magnitude, 1.0):
        eunit = eunit.units
    xname, yname = pretty_array_name(names["x"]), pretty_array_name(names["y"])
    ax.set_xlabel(f'{xname} [{eunit:~P}]')
    ax.set_ylabel(f'{yname} [{eunit:~P}]')

    ratio = (extent[1] - extent[0]) / (extent[3] - extent[2])
    if not max(ratio, 1 / ratio) > 10.0:
        ax.set_aspect('equal')

    ax.set(**ax_kwargs)

    if show_colorbar:
        divider = make_axes_locatable(ax)
        _kwargs = copy(colorbar_kwargs)
        position = _kwargs.pop('position', 'right')
        size = _kwargs.pop('size', '5%')
        pad = _kwargs.pop('pad', '2%')
        if position in ('top', 'bottom'):
            _kwargs.update({'orientation': 'horizontal'})
        cax = divider.append_axes(position=position, size=size, pad=pad)
        cbar = fig.colorbar(plot_object, cax, **_kwargs)

        if interp == 'projection':
            qunit = units['quantity'] * units['projection']
        elif interp == 'slice':
            qunit = units['quantity']
        if np.allclose(qunit.magnitude, 1.0):
            qunit = qunit.units
        qname = pretty_array_name(names["quantity"])
        cbar.set_label(f'{qname} [{qunit:~P}]')


def plot(
    snap: SnapLike,
    *,
    x: str = 'x',
    y: str = 'y',
    c: str = None,
    s: str = None,
    units: Dict[str, str] = None,
    xlim: Quantity = None,
    ylim: Quantity = None,
    ax: Any = None,
    ax_kwargs={},
    colorbar_kwargs={},
    **kwargs,
) -> Any:
    """Visualize SPH data as a particle plot.

    Visualize SPH data by plotting the particles, or a subset of
    the particles, possibly with marker colors and different sizes.

    Parameters
    ----------
    snap
        The Snap (or SubSnap) object to visualize.
    x
        The x-coordinate for the visualization. Must be a string to
        pass to Snap. Default is 'x'.
    y
        The y-coordinate for the visualization. Must be a string to
        pass to Snap. Default is 'y'.
    c
        The quantity to color the particles. Must be a string to
        pass to Snap.
    s
        The quantity to set the particle size. Must be a string to
        pass to Snap.
    units
        The units of the plot as a dictionary. The keys correspond to
        quantities such as 'position', 'density', 'velocity', and so on.
        The values are strings representing units, e.g. 'g/cm^3' for
        density.
    xlim
        The range in the x-coord as (xmin, xmax) where xmin/xmax can be
        floats or quantities with units of length.
    ylim
        The range in the y-coord as (ymin, ymax) where ymin/ymax can be
        floats or quantities with units of length.
    ax
        A matplotlib Axes handle.
    ax_kwargs
        Keyword arguments to pass to matplotlib Axes.
    colorbar_kwargs
        Keyword arguments to pass to matplotlib Colorbar.
    **kwargs
        Additional keyword arguments to pass to matplotlib
        functions.

    Returns
    -------
    ax
        The matplotlib Axes object.

    Examples
    --------
    Show the particles in xy-plane.

    >>> plonk.plot(snap=snap)

    Alternatively, access the function as a method on the Snap object.

    >>> snap.plot()

    Plot density against x.

    >>> snap.plot(x='x', y='density')

    Color particles by density.

    >>> snap.plot(x='x', y='y', c='density')

    Set units for the plot.

    >>> units = {'position': 'au', 'density': 'g/cm^3'}
    >>> snap.plot(x='x', y='y', c='density', units=units)
    """
    logger.debug(f'Plotting particles "{x}" vs "{y}"" on snap: {snap.file_path.name}')
    _kwargs = copy(kwargs)

    if ax is None:
        fig, ax = plt.subplots()
    else:
        fig = ax.figure

    if c is None and s is None:
        # If color (c) and size (s) are not required we color each
        # particle type differently
        try:
            subsnaps: Sequence[SnapLike] = snap.subsnaps_as_list(squeeze=False)
        except AttributeError:
            # Sinks do not have subsnaps
            subsnaps = [snap]
    else:
        # The subsnaps list is just a list with the original snap
        subsnaps = [snap]

    for subsnap in subsnaps:
        _x, _y, _c, _s, _units = _plot_data(
            snap=subsnap, x=x, y=y, c=c, s=s, units=units
        )
        _plot_plot(
            x=_x,
            y=_y,
            c=_c,
            s=_s,
            units=_units,
            xlim=xlim,
            ylim=ylim,
            names={'x': x, 'y': y, 'c': c, 's': s},
            fig=fig,
            ax=ax,
            ax_kwargs=ax_kwargs,
            colorbar_kwargs=colorbar_kwargs,
            **_kwargs,
        )

    return ax


def _plot_data(snap, x, y, c, s, units):
    _x: Quantity = snap[x]
    _y: Quantity = snap[y]
    _c: Quantity = snap[c] if c is not None else None
    _s: Quantity = snap[s] if s is not None else None

    _units = {
        'x': _get_unit(snap, x, units),
        'y': _get_unit(snap, y, units),
        'c': _get_unit(snap, c, units),
        's': _get_unit(snap, s, units),
    }

    _x = _x.to(_units['x']).magnitude
    _y = _y.to(_units['y']).magnitude
    if _c is not None:
        _c = _c.to(_units['c']).magnitude
    if _s is not None:
        _s = _s.to(_units['s']).magnitude

    try:
        # ignore accreted particles
        h: ndarray = snap['smoothing_length'].m
        mask = h > 0
        _x = _x[mask]
        _y = _y[mask]
        if _c is not None:
            _c = _c[mask]
        if _s is not None:
            _s = _s[mask]
    except KeyError:
        # sink particles do not have smoothing length
        pass

    if _s is not None:
        _s = 100 * _s / _s.max()

    return _x, _y, _c, _s, _units


def _plot_plot(
    x, y, c, s, units, xlim, ylim, names, fig, ax, ax_kwargs, colorbar_kwargs, **kwargs,
):
    show_colorbar = kwargs.pop('show_colorbar', c is not None)

    if s is None and c is None:
        plots.plot(x=x, y=y, ax=ax, **kwargs)
    else:
        plot_object = plots.scatter(x=x, y=y, c=c, s=s, ax=ax, **kwargs)

    ratio = (x.max() - x.min()) / (y.max() - y.min())
    if not max(ratio, 1 / ratio) > 10.0:
        ax.set_aspect('equal')

    xunit, yunit = units['x'], units['y']
    if np.allclose(xunit.magnitude, 1.0):
        xunit = xunit.units
    if np.allclose(yunit.magnitude, 1.0):
        yunit = yunit.units
    xname, yname = pretty_array_name(names["x"]), pretty_array_name(names["y"])
    ax.set_xlabel(f'{xname} [{xunit:~P}]')
    ax.set_ylabel(f'{yname} [{yunit:~P}]')

    ax.set(**ax_kwargs)

    if xlim is not None:
        if not isinstance(xlim, Quantity):
            _xlim = xlim * xunit
        else:
            _xlim = xlim.to(xunit)
        ax.set_xlim(_xlim.magnitude)
    if ylim is not None:
        if not isinstance(ylim, Quantity):
            _ylim = ylim * yunit
        else:
            _ylim = ylim.to(yunit)
        ax.set_ylim(_ylim.magnitude)

    if show_colorbar:
        divider = make_axes_locatable(ax)
        _kwargs = copy(colorbar_kwargs)
        position = _kwargs.pop('position', 'right')
        size = _kwargs.pop('size', '5%')
        pad = _kwargs.pop('pad', '2%')
        if position in ('top', 'bottom'):
            _kwargs.update({'orientation': 'horizontal'})
        cax = divider.append_axes(position=position, size=size, pad=pad)
        cbar = fig.colorbar(plot_object, cax, **_kwargs)

        cunit = units['c']
        if np.allclose(cunit.magnitude, 1.0):
            cunit = cunit.units
        cname = pretty_array_name(names["c"])
        cbar.set_label(f'{cname} [{cunit:~P}]')


def _convert_units_for_interpolation(
    *,
    snap: SnapLike,
    quantity: str,
    interpolated_data: ndarray,
    extent: ndarray,
    units: Dict[str, Any],
    interp: str,
):
    required_keys = {'extent', 'projection', 'quantity'}
    if not set(units) == required_keys:
        raise ValueError(f'units dictionary requires: {required_keys}')
    quantity_unit = snap.array_code_unit(quantity)
    if interp == 'projection':
        proj_unit = units['quantity'] * units['projection']
        data = (interpolated_data * quantity_unit * snap.code_units['length']).to(
            proj_unit.units
        ).magnitude / proj_unit.magnitude
    elif interp == 'slice':
        data = (interpolated_data * quantity_unit).to(
            units['quantity'].units
        ).magnitude / units['quantity'].magnitude

    new_extent = tuple(
        (extent * snap.code_units['length']).to(units['extent'].units).magnitude
        / units['extent'].magnitude
    )

    return data, new_extent


def _convert_units_for_cmap(vm, name, units, interp):
    if interp == 'projection':
        quantity_unit = units['quantity'] * units['projection']
    elif interp == 'slice':
        quantity_unit = units['quantity']
    if vm is not None:
        if isinstance(vm, Quantity):
            if vm.dimensionality != quantity_unit.dimensionality:
                raise ValueError(f'{name} has incorrect units')
            vm = (vm / quantity_unit).to_base_units().magnitude
        else:
            logger.warning(f'{name} has no units, assuming same units as quantity')

    return vm


def _get_unit(snap, name, units):
    if name is None:
        return None
    if name == 'projection':
        base_name = 'projection'
    else:
        base_name = snap.base_array_name(name)
    if units is not None:
        if name in units:
            return 1 * plonk_units(units[name])
        if base_name in units:
            return 1 * plonk_units(units[base_name])
    if base_name in snap.default_units:
        return 1 * plonk_units(snap.default_units[base_name])
    if name == 'projection':
        return 1 * snap['position'].units
    return 1 * snap[base_name].units



"""Snapshot data access and manipulation.

It contains Plonk the implementation of smoothed particle hydrodynamics
snapshot file.
"""

from .readers import load_snap
from .snap import Sinks, Snap, SnapLike, SubSnap
from .utils import gravitational_constant_in_code_units

__all__ = [
    'Sinks',
    'Snap',
    'SnapLike',
    'SubSnap',
    'gravitational_constant_in_code_units',
    'load_snap',
]



"""Testing Snap."""

from pathlib import Path

import numpy as np
import pytest
from scipy.spatial.transform import Rotation

import plonk

from .stubdata.phantom_snapshot import (
    array_name_map,
    mean_array_values,
    properties,
    std_array_values,
)

TEST_FILE = Path(__file__).parent / 'stubdata/phantom_00000.h5'
AVAILABLE_ARRAYS = [
    'angular_momentum',
    'angular_velocity',
    'azimuthal_angle',
    'density',
    'dust_to_gas_ratio',
    'eccentricity',
    'id',
    'inclination',
    'keplerian_frequency',
    'kinetic_energy',
    'mass',
    'momentum',
    'polar_angle',
    'position',
    'pressure',
    'radius_cylindrical',
    'radius_spherical',
    'semi_major_axis',
    'smoothing_length',
    'sound_speed',
    'specific_angular_momentum',
    'stokes_number',
    'stopping_time',
    'sub_type',
    'temperature',
    'timestep',
    'type',
    'velocity',
    'velocity_divergence',
    'velocity_radial_cylindrical',
    'velocity_radial_spherical',
]


def test_load_phantom_snap():
    """Testing reading Phantom HDF5 snapshots."""
    # Read from Path
    snap = plonk.load_snap(TEST_FILE)
    snap.close_file()
    # Read from str
    snap = plonk.load_snap(str(TEST_FILE))
    snap.close_file()
    # Not exists
    with pytest.raises(FileNotFoundError):
        plonk.load_snap('does_not_exist.h5')


def test_get_item():
    """Testing getting items from Snap."""
    snap = plonk.load_snap(TEST_FILE)

    position = snap['position']
    assert position.shape == (2000, 3)

    subsnap = snap['gas']
    assert type(subsnap) == plonk.snap.snap.SubSnap

    subsnap = snap[:10]
    assert type(subsnap) == plonk.snap.snap.SubSnap
    assert len(subsnap) == 10

    subsnap = snap[[0, 1, 2, 3, 4, 5, 6, 7, 8, 9]]
    assert type(subsnap) == plonk.snap.snap.SubSnap
    assert len(subsnap) == 10

    subsnap = snap[(0, 1, 2, 3, 4, 5, 6, 7, 8, 9)]
    assert type(subsnap) == plonk.snap.snap.SubSnap
    assert len(subsnap) == 10

    snap.close_file()


def test_read_particle_arrays_from_phantom():
    """Testing reading Phantom HDF5 snapshot particle arrays."""
    snap = plonk.load_snap(TEST_FILE)
    _check_arrays(snap)
    snap.close_file()


def test_read_properties_from_phantom():
    """Testing reading Phantom HDF5 snapshot properties."""
    snap = plonk.load_snap(TEST_FILE)

    for key, value in properties.items():
        if isinstance(snap.properties[key], plonk.units.Quantity):
            snap_value = snap.properties[key].magnitude
            numpy_array = True
        elif isinstance(snap.properties[key], str):
            snap_value = snap.properties[key]
            numpy_array = False
        else:
            snap_value = snap.properties[key]
            numpy_array = True
        if numpy_array:
            np.testing.assert_allclose(snap_value, value)
        else:
            assert snap_value == value

    snap.close_file()


def test_available_loaded_arrays():
    """Testing seeing available/loaded arrays on Snap."""
    snap = plonk.load_snap(TEST_FILE)

    assert snap.available_arrays() == AVAILABLE_ARRAYS

    for arr in [
        'position_x',
        'position_y',
        'position_z',
        'position_mag',
        'h',
        'dust_to_gas_ratio_tot',
    ]:
        snap[arr]

    assert snap.loaded_arrays() == ['dust_to_gas_ratio', 'position', 'smoothing_length']

    snap.close_file()


def test_array_code_unit():
    """Testing getting array code unit."""
    snap = plonk.load_snap(TEST_FILE)

    position_unit = 149600000000.0 * plonk.units('meter')
    assert snap.array_code_unit('position') == position_unit

    for arr in ['position', 'position_x', 'x']:
        snap.array_code_unit(arr)

    with pytest.raises(ValueError):
        snap.array_code_unit('does_not_exist')

    snap.close_file()


def test_rotate_snap():
    """Testing rotating Snap."""
    snap = plonk.load_snap(TEST_FILE)

    snap['position']
    snap['radius_cylindrical']
    snap.sinks['position']
    snap.rotate(axis=(1, 2, 3), angle=np.pi)
    snap.rotate(axis=(1, 2, 3), angle=-np.pi)
    _check_arrays(snap)

    snap.rotate(axis=(1, 2, 3), angle=np.pi)
    snap.reset()
    _check_arrays(snap)

    rot = np.array([1, 2, 3])
    rot = rot / np.linalg.norm(rot)
    rot *= np.pi
    rotation = Rotation.from_rotvec(rot)
    snap.rotate(rotation=rotation)
    _check_arrays(snap)

    snap.close_file()


def test_translate_snap():
    """Testing translating Snap."""
    snap = plonk.load_snap(TEST_FILE)

    snap['position']
    snap.sinks['position']
    snap.translate(translation=(100, 200, 300), unit='au')
    snap.translate(translation=(-100, -200, -300), unit='au')
    _check_arrays(snap)

    snap.translate(translation=(100, 200, 300), unit='au')
    snap.reset()
    _check_arrays(snap)

    with pytest.raises(ValueError):
        snap.translate(translation=(100, 200, 300))

    with pytest.raises(ValueError):
        snap.translate(translation=(100, 200))

    snap.translate(translation=(100, 200, 300) * plonk.units('au'), unit='au')
    snap.translate(translation=(-100, -200, -300) * plonk.units('au'), unit='au')
    _check_arrays(snap)

    snap.close_file()


def test_write_to_dataframe():
    """Testing writing Snap to DataFrame."""
    snap = plonk.load_snap(TEST_FILE)

    columns = ['position', 'density', 'smoothing_length']
    snap.to_dataframe(columns=columns)

    columns = ['position', 'density', 'smoothing_length']
    snap.to_dataframe(columns=columns, units=['au', 'g/cm^3', 'au'])

    snap.close_file()


def test_subsnap():
    """Testing getting SubSnap."""
    snap = plonk.load_snap(TEST_FILE)

    gas = snap['gas']
    assert len(gas) == 1000

    subsnap = snap[0:100]
    assert len(subsnap) == 100

    subsnap = snap[[0, 1, 2]]
    assert len(subsnap) == 3

    subsnap = snap[(0, 1, 2)]
    assert len(subsnap) == 3

    subsnap = snap[0]
    assert len(subsnap) == 1

    snap.close_file()


def test_sinks():
    """Testing getting sink particles."""
    snap = plonk.load_snap(TEST_FILE)

    sinks = snap.sinks

    assert snap.num_sinks == 1
    assert len(sinks) == 1

    snap.close_file()


def test_set_array():
    """Testing setting array on particles."""
    snap = plonk.load_snap(TEST_FILE)

    particle_array = np.arange(len(snap)) * plonk.units('dimensionless')
    snap['array'] = particle_array
    np.testing.assert_allclose(snap['array'].m, particle_array.m)

    sink_array = np.arange(len(snap.sinks)) * plonk.units('dimensionless')
    snap.sinks['array'] = sink_array
    np.testing.assert_allclose(snap.sinks['array'].m, sink_array.m)

    snap.close_file()


def test_bulk_load():
    """Testing bulk loading arrays."""
    snap = plonk.load_snap(TEST_FILE)
    snap.bulk_load()

    snap.close_file()


def test_read_write_extra():
    """Testing read write extra arrays."""
    snap = plonk.load_snap(TEST_FILE)

    filename = Path('tmp.h5')

    arr = np.arange(len(snap)) * plonk.units('dimensionless')
    snap['my_array'] = arr
    snap.write_extra_arrays(arrays=['my_array'], filename=filename)
    snap = None

    snap = plonk.load_snap(TEST_FILE)
    snap.read_extra_arrays(filename=filename)
    np.allclose(snap['my_array'], arr)

    filename.unlink()

    snap.close_file()


def test_plot_as_methods():
    """Testing plot methods."""
    snap = plonk.load_snap(TEST_FILE)

    snap.image('density', number_of_pixels=(16, 16))
    snap.plot()

    sinks = snap.sinks
    sinks.plot()

    subsnap = snap['gas']
    subsnap.image('density', number_of_pixels=(16, 16))
    subsnap.plot()

    snap.close_file()


def test_context():
    """Testing cache context manager."""
    snap = plonk.load_snap(TEST_FILE)

    snap.cache_arrays = True
    with snap.context(cache=False):
        assert snap.loaded_arrays() == []
        with snap.context(cache=True):
            snap.image('density', number_of_pixels=(16, 16))
            assert snap.loaded_arrays() == [
                'density',
                'mass',
                'position',
                'smoothing_length',
            ]
        assert snap.loaded_arrays() == []
    assert snap.loaded_arrays() == []

    snap.close_file()


def _check_arrays(snap):
    for array in mean_array_values.keys():
        np.testing.assert_allclose(
            snap.array_in_code_units(array_name_map[array]).mean(),
            mean_array_values[array],
        )

    for array in std_array_values.keys():
        np.testing.assert_allclose(
            snap.array_in_code_units(array_name_map[array]).std(),
            std_array_values[array],
        )



"""Testing evolution files."""

from pathlib import Path

import numpy as np
import pytest

import plonk

from .stubdata.phantom_evolution import mean_values

TEST_FILE = Path(__file__).parent / 'stubdata/phantom01.ev'


def test_read_evolution():
    """Test reading Phantom evolution files."""
    # Read from Path
    plonk.load_ev(TEST_FILE)

    # Read from str
    plonk.load_ev(str(TEST_FILE))

    # Not exists
    with pytest.raises(FileNotFoundError):
        plonk.load_ev('does_not_exist.ev')


def test_read_evolution_data():
    """Test reading data from Phantom evolution files."""
    ev = plonk.load_ev(TEST_FILE)

    assert set(ev.columns) == mean_values.keys()

    for key in ev.columns:
        np.testing.assert_allclose(ev[key].mean(), mean_values[key])



"""Logging."""

import logging
import platform

# Logging
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

console_handler = logging.StreamHandler()
file_handler = logging.FileHandler('.plonk.log')

console_handler.setLevel(logging.INFO)
file_handler.setLevel(logging.DEBUG)

console_format = logging.Formatter('%(levelname)s - %(message)s')
file_format = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
console_handler.setFormatter(console_format)
file_handler.setFormatter(file_format)

logger.addHandler(console_handler)
logger.addHandler(file_handler)


def get_os_info():
    """Get the operating system version for logging."""
    system = platform.system()
    if system == 'Darwin':
        system = 'macOS'
    release = platform.release()
    return f'{system} version: {release}'


def logger_init(__version__):
    """Log versions and platform."""
    logger.debug(f'Plonk v{__version__} on Python {platform.python_version()}')
    logger.debug(f'{get_os_info()}, {platform.machine()}')



"""Testing SPH neighbours and kernels."""

from pathlib import Path

import numpy as np
import pytest

import plonk
from plonk import analysis

TEST_FILE = Path(__file__).parent / 'stubdata/phantom_00000.h5'
NEIGHBOURS_FILE = Path(__file__).parent / 'stubdata/neighbours.csv'

with open(NEIGHBOURS_FILE) as fp:
    NEIGHBOURS = list()
    lines = fp.readlines()
    for line in lines:
        if line[0] == '#':
            continue
        NEIGHBOURS.append([int(item) for item in line.strip().split(',')])


def test_get_neighbours():
    """Testing getting particle neighbours."""
    snap = plonk.load_snap(TEST_FILE)

    with pytest.raises(ValueError):
        snap.set_kernel('kernel_does_not_exist')

    snap.set_kernel('cubic')

    subsnap = snap['gas']
    indices = np.arange(len(subsnap))
    neigh = subsnap.neighbours(indices)
    for idxi, idxj in enumerate(indices):
        assert set(neigh[idxi]) == set(NEIGHBOURS[idxj])

    snap.close_file()


def test_sph_derivative():
    """Test SPH derivative."""
    snap = plonk.load_snap(TEST_FILE)

    analysis.sph.derivative(
        snap=snap, derivative='div', quantity='velocity', kernel='cubic',
    )

    snap.close_file()



"""Reader for snapshot files."""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Union

from ..._logging import logger
from .phantom import generate_snap_from_file as read_phantom

if TYPE_CHECKING:
    from ..snap import Snap

_data_sources = ['Phantom']


def load_snap(
    filename: Union[str, Path],
    data_source: str = 'Phantom',
    config: Union[str, Path] = None,
) -> Snap:
    """Load a snapshot from file.

    Parameters
    ----------
    filename
        Path to snapshot file.
    data_source : optional
        The SPH software that produced the data. Default is 'Phantom'.
    config : optional
        The path to a Plonk config.toml file.

    Returns
    -------
    Snap
        The Snap object.
    """
    if data_source not in _data_sources:
        raise ValueError(
            f'Unknown data source. Available data sources:\n{_data_sources}'
        )

    if data_source == 'Phantom':
        try:
            return read_phantom(filename=filename, config=config)
        except FileNotFoundError as e:
            logger.error(f'File not found: {filename}')
            raise e
        except OSError as e:
            # Catch errors raised by h5py due to file corruption
            logger.error(f'File likely corrupted: {filename}')
            raise e
    raise RuntimeError('Cannot load snap')



"""Profiles to reduce data to 1-dimension."""

# Heavily inspired by pynbody (https://pynbody.github.io/).

from __future__ import annotations

from bisect import bisect
from copy import copy
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from numpy import ndarray
from pandas import DataFrame
from scipy.interpolate import interp1d

from .._logging import logger
from .._units import Quantity
from .._units import units as plonk_units
from ..utils.math import average
from ..utils.utils import is_documented_by, pretty_array_name
from ._profiles import extra_profiles

if TYPE_CHECKING:
    from ..snap.snap import SnapLike

_aggregations = ('average', 'mean', 'median', 'std', 'sum')


class Profile:
    """Profiles.

    A profile is a binning of particles in Cartesian slices, or
    cylindrical or spherical shells around the origin, i.e. (0, 0, 0).
    For cylindrical profiles the cylindrical cells are perpendicular to
    the xy-plane, i.e. the bins are averaged azimuthally and in the
    z-direction. Cartesian profiles can be in the x-, y-, and
    z-directions.

    Parameters
    ----------
    snap
        The Snap object.
    ndim : optional
        The dimension of the profile. For ndim == 2, the radial binning
        is cylindrical in the xy-plane. For ndim == 3, the radial
        binning is spherical. For ndim == 1, the radial binning is
        Cartesian along the x-axis. Default is 2.
    cmin : optional
        The minimum coordinate for binning. Can be a string, e.g.
        '10 au', or a quantity with units, e.g. plonk.units('10 au').
        Defaults to minimum on the particles.
    cmax : optional
        The maximum coordinate for binning. Can be a string, e.g.
        '10 au', or a quantity with units, e.g. plonk.units('10 au').
        Defaults to the 99 percentile distance.
    n_bins : optional
        The number of radial bins. Default is 100.
    aggregation : optional
        The method to aggregate particle quantities in bins by. Options
        are 'average', 'mean', or 'median'. Here 'average' is a
        mass-weighted average. Default is 'average'.
    spacing : optional
        The spacing of radial bins. Can be 'linear' or 'log'. Default is
        'linear'.
    coordinate : optional
        The coordinate ('x', 'y', or 'z') for Cartesian profiles only,
        i.e. when ndim==1. Default is 'x'. For cylindrical and spherical
        profiles the coordinate is 'radius'.
    ignore_accreted : optional
        Ignore particles accreted onto sinks. Default is True.

    Examples
    --------
    Generate profile from snapshot.

    >>> prof = plonk.load_profile(snap=snap)
    >>> prof = plonk.load_profile(snap=snap, n_bins=300)
    >>> prof = plonk.load_profile(snap=snap, cmin='10 au', cmax='300 au')
    >>> prof = plonk.load_profile(snap=snap, spacing='log')

    To access a profile.

    >>> prof['surface_density']
    >>> prof['scale_height']

    To set a new profile.

    >>> prof['aspect_ratio'] = prof['scale_height'] / prof['radius']

    Alternatively use the add_profile decorator.

    >>> @prof.add_profile
    ... def mass(prof):
    ...     M = prof.snap['mass']
    ...     return prof.particles_to_binned_quantity('sum', M)

    Plot one or many quantities on the profile.

    >>> prof.plot('radius', 'density')
    >>> prof.plot('radius', ['angular_momentum_x', 'angular_momentum_y'])

    Plot a quantity on the profile with units.

    >>> units = {'position': 'au', 'surface_density'='g/cm^2'}
    >>> prof.plot('radius', 'surface_density', units=units)
    """

    def __init__(
        self,
        snap: SnapLike,
        ndim: int = 2,
        cmin: Any = None,
        cmax: Any = None,
        n_bins: int = 100,
        aggregation: str = 'average',
        spacing: str = 'linear',
        coordinate: str = 'x',
        ignore_accreted: bool = True,
    ):

        self.snap = snap
        self.ndim = ndim
        self.aggregation = _check_aggregation(aggregation)
        self.spacing = _check_spacing(spacing)
        self.properties: Dict[str, Any] = {}

        self._profiles: Dict[str, Quantity] = {}
        self._profile_functions: Dict[str, Callable] = {}
        self._default_units = copy(self.snap._default_units)

        self._weights = self.snap['mass']
        self._mask = self._setup_particle_mask(ignore_accreted)
        self._x = self._calculate_x(coordinate)
        self.range = self._set_range(cmin, cmax)
        self.n_bins = n_bins

        self.bin_edges, self['size'] = self._setup_bins()
        self.bin_centers = 0.5 * (self.bin_edges[:-1] + self.bin_edges[1:])
        self._particle_bin = np.digitize(self._x.magnitude, self.bin_edges.magnitude)
        self.bin_indicies = self._set_particle_bin_indicies()

        if ndim == 1:
            self._coordinate = coordinate
        else:
            self._coordinate = 'radius'
        self._profiles[self._coordinate] = self.bin_centers
        self._profiles['number'] = np.histogram(
            self._x.magnitude, self.bin_edges.magnitude
        )[0] * plonk_units('dimensionless')

        try:
            num_separate_dust = len(self.snap.num_particles_of_type['dust'])
        except KeyError:
            num_separate_dust = 0
        num_mixture_dust_species = self.snap.num_dust_species - num_separate_dust

        # Add pre-defined profiles
        extra_profiles(self, num_mixture_dust_species)

    def add_profile(self, fn: Callable) -> Callable:
        """Decorate function to add profile to Profile.

        Parameters
        ----------
        fn
            A function that returns the profile as an array. The name of
            the function is the string with which to reference the array.

        Returns
        -------
        Callable
            The function which returns the array.
        """
        self._profile_functions[fn.__name__] = fn
        return fn

    def _setup_particle_mask(self, ignore_accreted: bool) -> ndarray:
        if ignore_accreted is False:
            return np.ones(len(self.snap), dtype=bool)
        h: Quantity = self.snap['h']
        return h > 0

    def _calculate_x(self, coordinate) -> Quantity:
        pos: Quantity = self.snap['position']
        pos = pos[self._mask]
        if self.ndim == 1:
            if coordinate == 'x':
                return pos[:, 0]
            if coordinate == 'y':
                return pos[:, 1]
            if coordinate == 'z':
                return pos[:, 2]
            raise ValueError('coordinate must be "x", "y", or "z" for ndim==1')
        if self.ndim == 2:
            return np.sqrt(pos[:, 0] ** 2 + pos[:, 1] ** 2)
        if self.ndim == 3:
            return np.sqrt(pos[:, 0] ** 2 + pos[:, 1] ** 2 + pos[:, 2] ** 2)
        raise ValueError('Unknown ndim: cannot calculate x array')

    def _set_range(self, cmin: Any, cmax: Any) -> Tuple[float, float]:
        if cmin is None:
            rmin = self._x.min()
        else:
            rmin = Quantity(cmin)
            if not rmin.dimensionality == Quantity('cm').dimensionality:
                raise ValueError('must specify cmin units, e.g. cmin="10 au"')
        if cmax is None:
            rmax = np.percentile(self._x.magnitude, 99, axis=0) * self._x.units
        else:
            rmax = Quantity(cmax)
            if not rmax.dimensionality == Quantity('cm').dimensionality:
                raise ValueError('must specify cmin units, e.g. cmax="100 au"')

        return rmin, rmax

    def _setup_bins(self) -> Quantity:
        bin_edges = self._bin_edges()
        if self.ndim == 1:
            bin_sizes = bin_edges[1:] - bin_edges[:-1]
        elif self.ndim == 2:
            bin_sizes = np.pi * (bin_edges[1:] ** 2 - bin_edges[:-1] ** 2)
        elif self.ndim == 3:
            bin_sizes = 4 / 3 * np.pi * (bin_edges[1:] ** 3 - bin_edges[:-1] ** 3)
        return bin_edges, bin_sizes

    def _bin_edges(self):
        if self.spacing == 'linear':
            bin_edges = (
                np.linspace(
                    self.range[0].magnitude, self.range[1].magnitude, self.n_bins + 1
                )
                * self.range[0].units
            )
        elif self.spacing == 'log':
            bin_edges = (
                np.logspace(
                    np.log10(self.range[0].magnitude),
                    np.log10(self.range[1].magnitude),
                    self.n_bins + 1,
                )
                * self.range[0].units
            )
        else:
            raise ValueError('Cannot determine spacing to setup bins')
        return bin_edges

    def _set_particle_bin_indicies(self) -> List[ndarray]:
        sortind = self._particle_bin.argsort()
        sort_pind = self._particle_bin[sortind]
        binind = list()
        prev_index = bisect(sort_pind, 0)
        for i in range(self.n_bins):
            new_index = bisect(sort_pind, i + 1)
            binind.append(np.sort(sortind[prev_index:new_index]))
            prev_index = new_index
        return binind

    def _getitem(self, name: str) -> Quantity:
        """Return the profile of a given kind."""
        name_root = '_'.join(name.split('_')[:-1])
        name_suffix = name.split('_')[-1]

        if name in self._profiles:
            return self._profiles[name]
        if name in self._profile_functions:
            self._profiles[name] = self._profile_functions[name](self)
            return self._profiles[name]

        if name_suffix in _aggregations:
            aggregation = name_suffix
            array_name = name_root
        else:
            aggregation = self.aggregation
            array_name = name
        try:
            array: Quantity = self.snap[array_name]
        except ValueError as e:
            logger.error(e)
            raise ValueError(f'array "{array_name}" not available on snap')
        if array.ndim == 1:
            self._profiles[name] = self.particles_to_binned_quantity(aggregation, array)
            return self._profiles[name]
        raise ValueError(
            'Requested profile has array dimension > 1.\nTo access x-, y-, or '
            'z-components, or magnitude of vector quantities,\ntry, for '
            'example, prof["velocity_x"] or prof["momentum_mag"].\nTo '
            'access dust profiles, try, for example, prof["stopping_time_001"] '
            'or\nprof["dust_mass_tot"].'
        )

    def __getitem__(self, name: str) -> Quantity:
        """Return the profile of a given kind."""
        return self._getitem(name)

    def __setitem__(self, name: str, item: Quantity):
        """Set the profile directly."""
        if not isinstance(item, Quantity):
            raise ValueError('"item" must be pint Quantity')
        if item.shape[0] != self.n_bins:
            raise ValueError('Length of array does not match number of bins')
        if name in self.loaded_profiles():
            raise ValueError(
                'Attempting to overwrite existing profile. To do so, first delete the '
                'profile\nwith del prof["profile"], then try again.'
            )
        if name in self.available_profiles():
            raise ValueError(
                'Attempting to set profile already available. '
                'See prof.available_profiles().'
            )
        self._profiles[name] = item

    def __delitem__(self, name):
        """Delete a profile from memory."""
        del self._profiles[name]

    def __len__(self):
        """Length as number of bins."""
        return self.n_bins

    def __repr__(self):
        """Dunder repr method."""
        return self.__str__()

    def __str__(self):
        """Dunder str method."""
        return f'<plonk.Profile "{self.snap.file_path.name}">'

    def _ipython_key_completions_(self):
        """Tab completion for IPython __getitem__ method."""
        return self.available_profiles()

    def loaded_profiles(self) -> List[str]:
        """Return a listing of loaded profiles."""
        return sorted(self._profiles.keys())

    def available_profiles(self) -> List[str]:
        """Return a listing of available profiles."""
        loaded = list(self.loaded_profiles())
        available = list(self._profile_functions.keys())
        snap_arrays = _1d_arrays(list(self.snap.available_arrays(verbose=True)))
        return sorted(set(loaded + available + snap_arrays))

    @property
    def default_units(self) -> Dict[str, Any]:
        """Profile default units."""
        return {
            key: self._default_units[key] for key in sorted(self._default_units.keys())
        }

    def set_units(self, **kwargs) -> Profile:
        """Set default unit for profiles.

        Parameters
        ----------
        kwargs
            Keyword arguments with keys as the profile name, e.g.
            'pressure', and with values as the unit as a string, e.g.
            'pascal'.

        Examples
        --------
        Set multiple default units.

        >>> profile.set_units(pressure='pascal', density='g/cm^3')
        """
        for key, val in kwargs.items():
            defaults = list(self.default_units) + list(self.snap.default_units)
            if key not in defaults:
                logger.info(f'adding profile {key} to default_units dict')
            self._default_units[key] = val

        return self

    def base_array_name(self, name: str) -> str:
        """Get the base array name from a string.

        For example, 'velocity_x' returns 'velocity', 'density' returns
        'density', 'dust_fraction_001' returns 'dust_fraction', 'x'
        returns 'position'.

        Parameters
        ----------
        name
            The name as a string

        Returns
        -------
        str
            The base name.
        """
        try:
            return self.snap.base_array_name(name)
        except ValueError:
            if name == self._coordinate:
                return 'position'
            return name

    def plot(
        self,
        x: str,
        y: Union[str, List[str]],
        units: Dict[str, Union[str, List[str]]] = None,
        std: str = None,
        ax: Any = None,
        ax_kwargs={},
        **kwargs,
    ) -> Any:
        """Plot profile.

        Parameters
        ----------
        x
            The x axis to plot as a string.
        y
            The y axis to plot. Can be string or multiple as a list of
            strings.
        units
            The units of the plot as a dictionary. The keys correspond to
            quantities such as 'position', 'density', 'velocity', and so on.
            The values are strings representing units, e.g. 'g/cm^3' for
            density.
        std: optional
            Add standard deviation on profile. Can be 'shading' or
            'errorbar'.
        ax : optional
            A matplotlib Axes object to plot to.
        ax_kwargs
            Keyword arguments to pass to matplotlib Axes.
        **kwargs
            Keyword arguments to pass to Axes plot method.

        Returns
        -------
        ax
            The matplotlib Axes object.
        """
        if std is not None:
            if std not in ('shading', 'errorbar'):
                raise ValueError('std must be "shading" or "errorbar"')

        if ax is None:
            _, ax = plt.subplots()

        if isinstance(y, str):
            y = [y]

        x_unit = _get_unit(self, x, units)
        y_unit = [_get_unit(self, yi, units) for yi in y]

        xdata = self[x].to(x_unit)

        xlabel = pretty_array_name(x)
        ax.set_xlabel(f'{xlabel} [{xdata.units:~P}]')

        for idx, yname in enumerate(y):
            ydata = self[yname].to(y_unit[idx])
            ylabel = pretty_array_name(yname)
            label = f'{ylabel} [{ydata.units:~P}]'
            if kwargs.get('label') is None:
                ax.plot(xdata.magnitude, ydata.magnitude, label=label, **kwargs)
            else:
                ax.plot(xdata.magnitude, ydata.magnitude, **kwargs)
            if std:
                _std_plot(self, xdata, ydata, yname, y_unit[idx], std, ax)
        ax.legend()
        ax.set(**ax_kwargs)

        return ax

    def to_function(self, profile: str, **kwargs) -> Callable:
        """Create function via interpolation.

        The function is of the coordinate of the profile, e.g.
        'radius', and returns values of the selected profile, e.g.
        'scale_height'. The function is generated from the
        scipy.interpolate function interp1d.

        Parameters
        ----------
        profile
            The profile function to create as a string, e.g.
            'scale_height'.

        Returns
        -------
        Callable
            The function.

        Examples
        --------
        Select all particles within a scale height in a disc.

        >>> scale_height = prof.to_function('scale_height')
        >>> subsnap = snap[np.abs(snap['z']) < scale_height(snap['R'])]
        """
        coord = self.bin_centers
        prof = self[profile]

        def fn(x):
            nonlocal coord, prof
            _coord = coord.to(x.units).magnitude
            _prof = prof.magnitude
            y = interp1d(_coord, _prof, fill_value='extrapolate', **kwargs)(x.magnitude)
            return y * self[profile].units

        return fn

    def to_dataframe(
        self, columns: List[str] = None, units: List[str] = None
    ) -> DataFrame:
        """Convert Profile to DataFrame.

        Parameters
        ----------
        columns : optional
            A list of columns to add to the data frame. If None, add all
            loaded columns. Default is None.
        units : optional
            A list of units corresponding to columns add to the data
            frame. Units must be strings, and must be base units. I.e.
            'cm' not '10 cm'. If None, use default, i.e. cgs. Default is
            None.

        Returns
        -------
        DataFrame
        """
        data = dict()
        if columns is None:
            columns = self.loaded_profiles()
        if units is None:
            _units = list()
            for column in columns:
                try:
                    _units.append(self[column].units)
                except AttributeError:
                    _units.append(plonk_units('dimensionless'))
        else:
            _units = list()
            for unit in units:
                u = plonk_units(unit)
                if np.allclose(u.m, 1.0):
                    _units.append(u.units)
                else:
                    raise ValueError(
                        'Units must be strings, and must be base units. '
                        'I.e. "cm" not "10 cm".'
                    )
        if len(_units) != len(columns):
            raise ValueError('units and columns must have same length')
        for column, unit in zip(columns, _units):
            try:
                name = column + f' [{unit:~}]'
                array = self[column].to(unit).magnitude
            except AttributeError:
                name = column
                array = self[column]
            data[name] = array
        return pd.DataFrame(data)

    def particles_to_binned_quantity(
        self, aggregation: str, array: Quantity
    ) -> Quantity:
        """Calculate binned quantities from particles.

        This takes care of the bin indices and ignoring accreted
        particles (if requested in instantiating the profile).

        Parameters
        ----------
        aggregation
            The aggregation function that acts on particles in the
            radial bin.
        array
            The particle array.
        """
        if aggregation not in _aggregations:
            raise ValueError('Cannot determine aggregation method')

        binned_quantity = np.zeros(self.n_bins)
        for idx, bin_ind in enumerate(self.bin_indicies):
            if bin_ind.size == 0:
                continue
            if aggregation == 'average':
                val = average(
                    array[self._mask][bin_ind],
                    weights=self._weights[self._mask][bin_ind],
                )
            elif aggregation == 'mean':
                val = np.mean(array[self._mask][bin_ind])
            elif aggregation == 'median':
                val = np.median(array[self._mask][bin_ind])
            elif aggregation == 'std':
                val = np.std(array[self._mask][bin_ind])
            elif aggregation == 'sum':
                val = np.sum(array[self._mask][bin_ind])

            binned_quantity[idx] = val.magnitude

        binned_quantity *= val.units

        return binned_quantity


@is_documented_by(Profile)
def load_profile(
    snap: SnapLike,
    ndim: int = 2,
    cmin: Any = None,
    cmax: Any = None,
    n_bins: int = 100,
    aggregation: str = 'average',
    spacing: str = 'linear',
    coordinate: str = 'x',
    ignore_accreted: bool = True,
) -> Profile:
    logger.debug(f'Loading profile: {snap.file_path.name}')
    return Profile(
        snap=snap,
        ndim=ndim,
        cmin=cmin,
        cmax=cmax,
        n_bins=n_bins,
        aggregation=aggregation,
        spacing=spacing,
        coordinate=coordinate,
        ignore_accreted=ignore_accreted,
    )


def _check_aggregation(method: str) -> str:
    if method in _aggregations:
        return method
    raise ValueError(
        f'Cannot determine aggregation method: choose from {_aggregations}'
    )


def _check_spacing(spacing: str) -> str:
    if spacing.lower() in ('lin', 'linear'):
        return 'linear'
    if spacing.lower() in ('log', 'logarithm', 'logarithmic'):
        return 'log'
    raise ValueError('Cannot determine spacing')


def _1d_arrays(arrays: list) -> list:
    _arrays = list()
    for array in arrays:
        if (array + '_x' in arrays) or (array + '_001' in arrays):
            pass
        else:
            _arrays.append(array)
    return _arrays


def _get_unit(profile, name, units):
    if name is None:
        return None
    base_name = profile.base_array_name(name)
    if units is not None:
        if name in units:
            return 1 * plonk_units(units[name])
        if base_name in units:
            return 1 * plonk_units(units[base_name])
    if base_name in profile.default_units:
        return 1 * plonk_units(profile.default_units[base_name])
    return 1 * profile[base_name].units


def _std_plot(profile, xdata, ydata, yname, yunit, std, ax):
    if yname.split('_')[-1] in _aggregations:
        _yname = '_'.join(yname.split('_')[:-1])
    else:
        _yname = yname
    try:
        y_std = profile[_yname + '_std']
    except ValueError:
        logger.warning('Cannot calculate standard deviation')
        return
    if profile.aggregation in ('std', 'sum'):
        y_mean = profile[_yname + '_mean']
    else:
        y_mean = ydata
    y_std = y_std.to(yunit).magnitude
    y_mean = y_mean.to(yunit).magnitude
    color = ax.lines[0].get_color()
    xdata = xdata.magnitude
    if std == 'shading':
        ax.fill_between(
            xdata, y_mean - y_std, y_mean + y_std, color=color, alpha=0.2,
        )
    elif std == 'errorbar':
        ax.errorbar(xdata, y_mean, yerr=y_std, linestyle='', color=color, alpha=0.5)



"""SPH kernels."""

import numba
import numpy as np


@numba.njit
def kernel_cubic(q):
    """Cubic kernel function.

    The form of this function includes the "C_norm" factor. I.e.
    C_norm * f(q).

    Parameters
    ----------
    q
        The particle separation in units of smoothing length, i.e. r/h.

    Returns
    -------
    float
        C_norm * f(q) for the cubic kernel.
    """
    if q < 1:
        return (0.75 * q ** 3 - 1.5 * q ** 2 + 1) / np.pi
    elif q < 2:
        return (-0.25 * (q - 2) ** 3) / np.pi
    else:
        return 0.0


@numba.njit
def kernel_gradient_cubic(q):
    """Cubic kernel gradient function.

    The form of this function includes the "C_norm" factor. I.e.
    C_norm * f'(q).

    Parameters
    ----------
    q
        The particle separation in units of smoothing length, i.e. r/h.

    Returns
    -------
    float
        C_norm * f'(q) for the cubic kernel.
    """
    if q < 1:
        return q * (2.25 * q - 3.0) / np.pi
    elif q < 2:
        return -0.75 * (q - 2.0) ** 2 / np.pi
    else:
        return 0.0


@numba.njit
def kernel_quintic(q):
    """Quintic kernel function.

    The form of this function includes the "C_norm" factor. I.e.
    C_norm * f(q).

    Parameters
    ----------
    q
        The particle separation in units of smoothing length, i.e. r/h.

    Returns
    -------
    float
        C_norm * f(q) for the quintic kernel.
    """
    if q < 1:
        return (-10 * q ** 5 + 30 * q ** 4 - 60 * q ** 2 + 66) / (120 * np.pi)
    elif q < 2:
        return (-((q - 3) ** 5) + 6 * (q - 2) ** 5) / (120 * np.pi)
    elif q < 3:
        return (-((q - 3) ** 5)) / (120 * np.pi)
    else:
        return 0.0


@numba.njit
def kernel_gradient_quintic(q):
    """Quintic kernel gradient function.

    The form of this function includes the "C_norm" factor. I.e.
    C_norm * f'(q).

    Parameters
    ----------
    q
        The particle separation in units of smoothing length, i.e. r/h.

    Returns
    -------
    float
        C_norm * f'(q) for the quintic kernel.
    """
    if q < 1:
        return (q * (-50 * q ** 3 + 120 * q ** 2 - 120)) / (120 * np.pi)
    elif q < 2:
        return (-5 * (q - 3) ** 4 + 30 * (q - 2.0) ** 4) / (120 * np.pi)
    elif q < 3:
        return (-5 * (q - 3) ** 4) / (120 * np.pi)
    else:
        return 0.0


@numba.njit
def kernel_wendland_c4(q):
    """Wendland C4 kernel function.

    The form of this function includes the "C_norm" factor. I.e.
    C_norm * f(q).

    Parameters
    ----------
    q
        The particle separation in units of smoothing length, i.e. r/h.

    Returns
    -------
    float
        C_norm * f(q) for the Wendland C4 kernel.
    """
    if q < 2:
        return (
            ((-q / 2 + 1) ** 6 * (35 * q ** 2 / 12 + 3 * q + 1)) * 495 / (256 * np.pi)
        )
    else:
        return 0.0


@numba.njit
def kernel_gradient_wendland_c4(q):
    """Wendland C4 kernel gradient function.

    The form of this function includes the "C_norm" factor. I.e.
    C_norm * f'(q).

    Parameters
    ----------
    q
        The particle separation in units of smoothing length, i.e. r/h.

    Returns
    -------
    float
        C_norm * f'(q) for the Wendland C4 kernel.
    """
    if q < 2:
        return (
            (
                11.6666666666667 * q ** 2 * (0.5 * q - 1) ** 5
                + 4.66666666666667 * q * (0.5 * q - 1) ** 5
            )
            * 495
            / (256 * np.pi)
        )
    else:
        return 0.0


kernel_names = (
    'cubic',
    'quintic',
    'Wendland C4',
)

kernel_radius = {
    'cubic': 2.0,
    'quintic': 3.0,
    'Wendland C4': 2.0,
}

kernel_function = {
    'cubic': kernel_cubic,
    'quintic': kernel_quintic,
    'Wendland C4': kernel_wendland_c4,
}

kernel_gradient_function = {
    'cubic': kernel_gradient_cubic,
    'quintic': kernel_gradient_quintic,
    'Wendland C4': kernel_gradient_wendland_c4,
}



"""Testing visualization."""

from pathlib import Path

import plonk
from plonk import visualize

AU = plonk.units('au')
TEST_FILE = Path(__file__).parent / 'stubdata/phantom_00000.h5'


def test_plot():
    """Test particle plot."""
    snap = plonk.load_snap(TEST_FILE)
    plonk.plot(snap=snap)

    snap.close_file()


def test_plot_with_kwargs():
    """Test particle plot with kwargs."""
    snap = plonk.load_snap(TEST_FILE)
    plonk.plot(
        snap=snap,
        x='x',
        y='y',
        c='density',
        units={'position': 'au', 'density': 'g/cm^3', 'projection': 'cm'},
    )

    snap.close_file()


def test_image_projection():
    """Test image projection."""
    snap = plonk.load_snap(TEST_FILE)
    plonk.image(snap=snap, quantity='density', number_of_pixels=(32, 32))

    snap.close_file()


def test_image_projection_with_kwargs():
    """Test image projection with kwargs."""
    snap = plonk.load_snap(TEST_FILE)
    plonk.image(
        snap=snap,
        quantity='density',
        x='x',
        y='y',
        units={'position': 'au', 'density': 'g/cm^3', 'projection': 'cm'},
        extent=(-150, 150, -150, 150) * AU,
        norm='linear',
        cmap='gist_heat',
        number_of_pixels=(32, 32),
    )

    snap.close_file()


def test_image_on_snap():
    """Test image projection as method on Snap."""
    snap = plonk.load_snap(TEST_FILE)
    snap.image(quantity='density', number_of_pixels=(32, 32))

    snap.close_file()


def test_image_slice():
    """Test image slice."""
    snap = plonk.load_snap(TEST_FILE)
    plonk.image(
        snap=snap, quantity='density', interp='slice', number_of_pixels=(32, 32)
    )

    snap.close_file()


def test_image_slice_with_kwargs():
    """Test image slice with kwargs."""
    snap = plonk.load_snap(TEST_FILE)
    plonk.image(
        snap=snap,
        quantity='density',
        interp='slice',
        x='x',
        y='y',
        units={'position': 'au', 'density': 'g/cm^3', 'projection': 'cm'},
        extent=(-150, 150, -150, 150) * AU,
        norm='linear',
        cmap='gist_heat',
        number_of_pixels=(32, 32),
    )

    snap.close_file()


def test_plot_smoothing_length():
    """Test plot smoothing length as circle."""
    snap = plonk.load_snap(TEST_FILE)

    visualize.plot_smoothing_length(snap=snap, indices=[0, 1])

    snap.close_file()


def test_get_extent():
    """Test getting extent from percentile."""
    snap = plonk.load_snap(TEST_FILE)

    visualize.get_extent_from_percentile(snap=snap, x='x', y='y')

    snap.close_file()



"""Visualize a simulation."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict

import matplotlib.pyplot as plt
import numpy as np
from numpy import ndarray

from .._logging import logger
from .visualization import image, plot, vector

if TYPE_CHECKING:
    from ..simulation.simulation import Simulation

KINDS = {'image': image, 'particle': plot, 'vector': vector}


class VisualizeSimulation:
    """Visualize a simulation.

    Parameters
    ----------
    sim
        The Simulation object.
    kind
        A string in 'image', 'particle', or 'vector' which indicates the
        plot kind.
    **kwargs
        Keyword arguments to pass to the plot method.

    Examples
    --------
    Visualize a simulation by density projection images.

    >>> viz = visualize_sim(sim=sim, kind='image', quantity='density')

    Alternatively.

    >>> sim.visualize(kind='image', quantity='density')

    Go forwards and backwards through snaps.

    >>> viz.next()
    >>> viz.prev()

    Go to a particular snap, or skip ahead.

    >>> viz.goto(10)
    >>> viz.next(5)
    """

    def __init__(self, sim: Simulation, kind: str, **kwargs):
        if kind not in KINDS:
            raise ValueError(f'kind must be one of {KINDS.keys()}')

        self.sim = sim
        self.kwargs: Dict[str, Any] = kwargs
        self.ax: Any = None
        self.snaps = sim.snaps

        self._particle_ids: ndarray = None
        self._kind = kind
        self._len = -1
        self._where = 0
        self._no_new_arrays = True

        self._plotting_function(kind=self.kind, idx=0)

    @property
    def kind(self):
        """Plot kind."""
        return self._kind

    @property
    def particle_ids(self):
        """Particle ids of snaps."""
        return self._particle_ids

    @particle_ids.setter
    def particle_ids(self, value: ndarray):
        subsnaps = [snap[value] for snap in self.snaps]
        self.snaps = subsnaps
        self._particle_ids = value

    @property
    def no_new_arrays(self):
        """Do not allow new arrays on Snap objects."""
        return self._no_new_arrays

    @no_new_arrays.setter
    def no_new_arrays(self, value):
        if isinstance(value, bool):
            self._no_new_arrays = value
        raise ValueError('no_new_arrays must be True or False')

    def _plotting_function(self, kind: str, idx: int):
        if self.ax is None:
            _, self.ax = plt.subplots()
        try:
            im = self.ax.images[0]
            cbar = im.colorbar
            cbar.remove()
        except (IndexError, AttributeError):
            pass
        self.ax.clear()

        snap = self.snaps[idx]
        loaded = set(snap.loaded_arrays())

        KINDS[kind](snap=snap, ax=self.ax, **self.kwargs)  # type: ignore

        new_arrays = set(snap.loaded_arrays()).symmetric_difference(loaded)
        if self._no_new_arrays:
            for array in new_arrays:
                del snap[array]

    def next(self, number: int = 1):
        """Visualize next snap."""
        idx = self._where + number
        if idx < len(self):
            self._plotting_function(kind=self.kind, idx=idx)
            self._where += number
        else:
            logger.info('Too far forward. Going to last snap.')
            self._plotting_function(kind=self.kind, idx=len(self) - 1)
            self._where = len(self) - 1

    def prev(self, number: int = 1):
        """Visualize previous snap."""
        idx = self._where - number
        if idx > 0:
            self._plotting_function(kind=self.kind, idx=idx)
            self._where -= number
        else:
            logger.info('Too far back. Going to first snap.')
            self._plotting_function(kind=self.kind, idx=0)
            self._where = 0

    def goto(self, idx: int):
        """Visualize particular snap by index."""
        if -len(self) < idx < len(self) - 1:
            self._plotting_function(kind=self.kind, idx=idx)
            self._where = np.mod(idx, len(self))
        else:
            raise ValueError('out of range')

    @property
    def index(self):
        """Get current snap index."""
        return self._where

    def __len__(self):
        """Length as number of snaps."""
        if self._len == -1:
            self._len = len(self.snaps)
        return self._len


def visualize_sim(sim: Simulation, kind: str, **kwargs) -> VisualizeSimulation:
    """Visualize a simulation.

    Parameters
    ----------
    sim
        The Simulation object.
    kind
        The kind of plot: 'particle', 'image', 'vector'.
    **kwargs
        Keyword arguments to pass to plotting methods such as
        plonk.image, plonk.plot, and plonk.vector.

    Returns
    -------
    VisualizeSimulation

    Examples
    --------
    Visualize a simulation by density projection images.

    >>> viz = visualize_sim(sim=sim, kind='image', quantity='density')

    Alternatively.

    >>> sim.visualize(kind='image', quantity='density')

    Go forwards and backwards through snaps.

    >>> viz.next()
    >>> viz.prev()

    Go to a particular snap, or skip ahead.

    >>> viz.goto(10)
    >>> viz.next(5)
    """
    return VisualizeSimulation(sim=sim, kind=kind, **kwargs)



"""Snap, SubSnap, Sinks classes for snapshot files.

The Snap class contains all information related to a smoothed particle
hydrodynamics simulation snapshot file. The SubSnap class is for
accessing a subset of particles in a Snap.
"""

from __future__ import annotations

from pathlib import Path
from typing import Any, Callable, Dict, List, Tuple, Union

import h5py
import numpy as np
import pandas as pd
from numpy import ndarray
from pandas import DataFrame
from scipy.spatial import cKDTree
from scipy.spatial.transform import Rotation

from .. import visualize
from .._logging import logger
from .._units import Quantity
from .._units import units as plonk_units
from ..utils.kernels import kernel_names, kernel_radius
from ..utils.math import norm
from . import context


class Snap:
    """Smoothed particle hydrodynamics Snap object.

    Snapshot files contain the state of the simulation at a point in
    time. Typical minimum data from a smoothed particle hydrodynamics
    simulation include the particle positions and smoothing length, from
    which the density field can be reconstructed, as well as the
    particle type. In addition, the particle velocities are required to
    restart the simulation.

    Other data stored in the snapshot file include equation of state,
    dust, and magnetic field information, as well as numerical
    quantities related to time-stepping.

    Examples
    --------
    Use load_snap to generate a Snap object.

    >>> snap = plonk.load_snap('file_name.h5')

    To access arrays on the particles.

    >>> snap['position']
    >>> snap['density']

    To access sink arrays.

    >>> snap.sinks['position']
    >>> snap.sinks['spin']

    To access a subset of particles as a SubSnap.

    >>> subsnap = snap[:100]
    >>> subsnap = snap[snap['x'] > 0]
    >>> subsnap = snap['gas']

    To set a new array directly.

    >>> snap['my_r'] = np.sqrt(snap['x'] ** 2 + snap['y'] ** 2)

    Alternatively, define a function.

    >>> @snap.add_array()
    ... def my_radius(snap):
    ...     return np.hypot(snap['x'], snap['y'])

    Or, use an existing one function.

    >>> @snap.add_array()
    ... def my_R(snap):
    ...     return plonk.analysis.particles.radial_distance(snap)
    """

    _array_aliases: Dict[str, str] = {}

    _vector_arrays = {
        'magnetic_field',
        'position',
        'spin',
        'velocity',
        'vorticity',
    }

    _dust_arrays = {
        'stopping_time',
        'dust_fraction',
        'dust_to_gas_ratio',
    }

    particle_type = {
        'gas': 1,
        'dust': 2,
        'boundary': 3,
        'star': 4,
        'darkmatter': 5,
        'bulge': 6,
    }

    def __init__(self):

        self.data_source = None
        self.file_path = None
        self._code_units = {}
        self._default_units = {}
        self._properties = {}
        self._array_code_units = {}
        self._array_registry: Dict[str, Callable] = {}
        self._sink_registry: Dict[str, Callable] = {}
        self._name_map = {}
        self._cache_arrays = True
        self._arrays = {}
        self._sink_arrays = {}
        self._sinks = None
        self._file_pointer = None
        self._num_particles = -1
        self._num_particles_of_type = -1
        self._num_sinks = -1
        self._num_dust_species = -1
        self.rotation = None
        self.translation = None
        self._tree = None

    def close_file(self):
        """Close access to underlying file."""
        self._file_pointer.close()

    def reopen_file(self):
        """Re-open access to the underlying file."""
        self._file_pointer = h5py.File(self.file_path, mode='r')

    def add_array(self, vector: bool = False, dust: bool = False) -> Callable:
        """Decorate function to add array to Snap.

        This function decorates a function that returns an array. The
        name of the function is the string with which to reference the
        array.

        The function being decorated should return a Pint Quantity
        array, not a unitless numpy array.

        Parameters
        ----------
        vector
            A bool to represent if the array is a vector in space that
            should have rotations applied to it. If True the rotation
            should be applied. If False the rotation cannot be applied.
            Default is False.
        dust
            A bool to represent if the array is a dust array, in that
            it has one column per dust species. Default is False.

        Returns
        -------
        Callable
            The decorator which returns the array.
        """

        def _add_array(fn):
            self._array_registry[fn.__name__] = fn
            if vector is True:
                self._vector_arrays.add(fn.__name__)
            if dust is True:
                self._dust_arrays.add(fn.__name__)
            return fn

        return _add_array

    def add_alias(self, name: str, alias: str) -> None:
        """Add alias to array.

        Parameters
        ----------
        name
            The name of the array.
        alias
            The alias to reference the array.
        """
        self._array_aliases[alias] = name

    def add_unit(self, name: str, unit: str) -> Snap:
        """Add missing code unit to array.

        Add code units to an array from file that has not had units set
        automatically.

        Parameters
        ----------
        name
            The name of the array
        unit
            A unit string representing the array code unit, e.g.
            '1.234 kg * m / s'.
        """
        self._array_code_units[name] = plonk_units(unit)
        if name in self._arrays:
            del self[name]

        return self

    def loaded_arrays(self) -> List[str]:
        """Return a list of loaded arrays.

        Returns
        -------
        List
            A list of names of loaded particle arrays.
        """
        return sorted(self._arrays.keys())

    def _available_arrays(
        self, sinks: bool = False, verbose: bool = False, aliases: bool = False
    ) -> List[str]:
        """Return a list of available arrays.

        Parameters
        ----------
        sinks
            If True, return available sink arrays. Default is
            False.
        verbose
            Also display suffixed arrays, e.g. 'position_x',
            'position_y', etc. Default is False
        aliases
            If True, return array aliases. Default is False.

        Returns
        -------
        List
            A list of names of available arrays.
        """
        if sinks:
            loaded = self.sinks.loaded_arrays()
            registered = list(self._sink_registry.keys())
        else:
            loaded = self.loaded_arrays()
            registered = list(self._array_registry.keys())
        if verbose:
            for arr in self._vector_arrays:
                if arr in registered:
                    registered += [f'{arr}_x', f'{arr}_y', f'{arr}_z', f'{arr}_mag']
            for arr in self._dust_arrays:
                if arr in registered:
                    registered += [
                        f'{arr}_{n:03}' for n in range(1, self.num_dust_species + 1)
                    ]
                    registered.append(f'{arr}_tot')

        if aliases:
            extra = [
                key
                for key, val in self._array_aliases.items()
                if val[0] in self.loaded_arrays() or val in self._array_registry
            ]
            if verbose:
                _extra = list()
                for arr in extra:
                    if self.base_array_name(arr) in self._vector_arrays:
                        for suffix in ['x', 'y', 'z', 'mag']:
                            _extra.append(arr + f'_{suffix}')
                    if self.base_array_name(arr) in self._dust_arrays:
                        suffixes = [f'{n+1:03}' for n in range(self.num_dust_species)]
                        suffixes += ['tot']
                        for suffix in suffixes:
                            _extra.append(arr + f'_{suffix}')
                extra += _extra
            return sorted(set(extra), key=lambda x: x.lower())

        return sorted(set(loaded + registered))

    def available_arrays(
        self, verbose: bool = False, aliases: bool = False
    ) -> List[str]:
        """Return a list of available particle arrays.

        Parameters
        ----------
        verbose
            Also display suffixed arrays, e.g. 'position_x',
            'position_y', etc. Default is False
        aliases
            If True, return array aliases. Default is False.

        Returns
        -------
        List
            A list of names of available arrays.
        """
        return self._available_arrays(sinks=False, verbose=verbose, aliases=aliases)

    def bulk_load(self, arrays: List[str] = None) -> Snap:
        """Load arrays into memory in bulk.

        Parameters
        ----------
        arrays
            A list of arrays to load as strings. If None, then load all
            available arrays.
        """
        if arrays is None:
            _arrays = self.available_arrays()
        else:
            _arrays = arrays
        with self.context(cache=True):
            for array in _arrays:
                try:
                    self[array]
                except ValueError as e:
                    logger.warning(f'Cannot load {array}\n{e}')

        return self

    def bulk_unload(self, arrays: List[str] = None) -> Snap:
        """Un-load arrays from memory in bulk.

        Parameters
        ----------
        arrays
            A list of arrays to load as strings. If None, then unload
            all loaded arrays.
        """
        if arrays is None:
            _arrays = self.loaded_arrays()
        else:
            _arrays = arrays
        for array in _arrays:
            try:
                del self[array]
            except KeyError:
                logger.warning(f'Cannot un-load {array}')

        return self

    @property
    def properties(self) -> Dict[str, Any]:
        """Snap properties."""
        return {key: self._properties[key] for key in sorted(self._properties.keys())}

    @property
    def default_units(self) -> Dict[str, Any]:
        """Snap default units."""
        return {
            key: self._default_units[key] for key in sorted(self._default_units.keys())
        }

    def set_units(self, **kwargs) -> Snap:
        """Set default unit for arrays.

        Parameters
        ----------
        kwargs
            Keyword arguments with keys as the array name, e.g.
            'pressure', and with values as the unit as a string, e.g.
            'pascal'.

        Examples
        --------
        Set multiple default units.

        >>> snap.set_units(pressure='pascal', density='g/cm^3')
        """
        for key, val in kwargs.items():
            if key not in self.default_units:
                logger.info(f'adding array {key} to default_units dict')
            self._default_units[key] = val

        return self

    @property
    def code_units(self) -> Dict[str, Any]:
        """Snap code units."""
        return {key: self._code_units[key] for key in sorted(self._code_units.keys())}

    @property
    def sinks(self) -> Sinks:
        """Sink particle arrays."""
        if self._sinks is None:
            self._sinks = Sinks(self)
        return self._sinks

    @property
    def num_particles(self) -> int:
        """Return number of particles."""
        if self._num_particles == -1:
            self._num_particles = len(self['type'])
        return self._num_particles

    @property
    def num_particles_of_type(self) -> Dict[str, Any]:
        """Return number of particles per type."""
        if self._num_particles_of_type == -1:
            int_to_name = {idx: name for name, idx in self.particle_type.items()}
            d = {}
            ptype: Quantity = self['type']
            stype: Quantity = self['sub_type']
            for idx, num in enumerate(np.bincount(ptype.magnitude)):
                if num > 0:
                    if idx == self.particle_type['dust']:
                        # Dust particle sub-type skips zero: 1, 2, 3 ...
                        d['dust'] = list(
                            np.bincount(stype[ptype.magnitude == idx].magnitude)[1:]
                        )
                    elif idx == self.particle_type['boundary']:
                        # Boundary particle sub-type: 0 (gas), 1, 2, 3... (dust)
                        d['boundary'] = list(np.bincount(stype[ptype == idx].magnitude))
                    else:
                        d[int_to_name[idx]] = num
            self._num_particles_of_type = d
        return self._num_particles_of_type

    @property
    def num_sinks(self) -> int:
        """Return number of sinks."""
        if self._num_sinks == -1:
            try:
                self._num_sinks = len(self._sink_registry['mass'](self))
            except KeyError:
                self._num_sinks = 0
        return self._num_sinks

    @property
    def num_dust_species(self) -> int:
        """Return number of dust species."""
        if self._num_dust_species == -1:
            self._num_dust_species = len(self._properties.get('grain_size', []))
        return self._num_dust_species

    @property
    def cache_arrays(self) -> bool:
        """Cache arrays in memory for faster access."""
        return self._cache_arrays

    @cache_arrays.setter
    def cache_arrays(self, value):
        self._cache_arrays = value

    def reset(
        self, arrays: bool = False, rotation: bool = True, translation: bool = True
    ) -> Snap:
        """Reset Snap.

        Reset rotation and translations transformations on the Snap to
        initial (on-file) values. In addition, unload cached arrays.

        Parameters
        ----------
        arrays
            Set to True to unload arrays from memory. Default is False.
        rotation
            Set to True to reset rotation. Default is True.
        translation
            Set to True to reset translation. Default is True.

        Returns
        -------
        Snap
            The reset Snap. Note that the reset operation is in-place.
        """
        if any((arrays, rotation, translation)):
            for arr in self.loaded_arrays():
                del self._arrays[arr]
            if self.num_sinks > 0:
                for arr in self.sinks.loaded_arrays():
                    del self._sink_arrays[arr]
        else:
            logger.warning('Select something to reset')

        if rotation:
            self.rotation = None
        if translation:
            self.translation = None

        return self

    def rotate(
        self,
        rotation: Rotation = None,
        axis: Union[ndarray, List[float], Tuple[float, float, float]] = None,
        angle: float = None,
    ) -> Snap:
        """Rotate snapshot.

        The rotation can be defined by a scipy Rotation object, or a
        combination of a vector, specifying a rotation axis, and an
        angle, specifying the rotation in radians.

        Parameters
        ----------
        rotation
            The rotation as a scipy.spatial.transform.Rotation object.
        axis
            An array specifying a rotation axis, like (x, y, z).
        angle
            A float specifying the rotation in radians.

        Returns
        -------
        Snap
            The rotated Snap. Note that the rotation operation is
            in-place.

        Examples
        --------
        Rotate a Snap by π/3 around [1, 1, 0].

        >>> axis = (1, 1, 0)
        >>> angle = np.pi / 3
        >>> snap.rotate(axis=axis, angle=angle)
        """
        logger.debug(f'Rotating snapshot: {self.file_path.name}')
        if rotation is not None:
            if axis is not None or angle is not None:
                logger.warning('ignoring axis and angle as rotation is passed in')
            _rotation = rotation
        else:
            _rotation = axis / norm(axis) * angle
        if isinstance(_rotation, (list, tuple, ndarray)):
            _rotation = Rotation.from_rotvec(_rotation)
        for arr in self._vector_arrays:
            if arr in self.loaded_arrays():
                array_m, array_u = self._arrays[arr].magnitude, self._arrays[arr].units
                self._arrays[arr] = _rotation.apply(array_m) * array_u
            if self.num_sinks > 0 and arr in self.sinks.loaded_arrays():
                array_m = self._sink_arrays[arr].magnitude
                array_u = self._sink_arrays[arr].units
                self._sink_arrays[arr] = _rotation.apply(array_m) * array_u
        for arr in self._vector_arrays:
            if arr in self.loaded_arrays():
                del self._arrays[arr]
            if self.num_sinks > 0 and arr in self.sinks.loaded_arrays():
                del self._sink_arrays[arr]

        if self.rotation is None:
            self.rotation = _rotation
        else:
            rot = _rotation * self.rotation
            self.rotation = rot

        return self

    def translate(
        self, translation: Union[Quantity, ndarray], unit: str = None
    ) -> Snap:
        """Translate snapshot.

        I.e. shift the snapshot origin.

        Parameters
        ----------
        translation
            The translation as an array like (x, y, z), optionally with
            units. If no units are specified you must specify the units
            parameter below.
        unit
            The length unit for the translation. E.g. 'au'.

        Returns
        -------
        Snap
            The translated Snap. Note that the translation operation is
            in-place.
        """
        logger.debug(f'Translating snapshot: {self.file_path.name}')
        if isinstance(translation, (list, tuple)):
            translation = np.array(translation)
        if translation.shape != (3,):
            raise ValueError('translation must be like (x, y, z)')
        if isinstance(translation, Quantity):
            if unit is not None:
                logger.warning('units argument ignored as translation has units')
        else:
            if unit is None:
                raise ValueError(
                    'translation must have units, or you must specify units argument'
                )
            translation *= plonk_units(unit)
        if 'position' in self.loaded_arrays():
            self._arrays['position'] += translation
        if self.num_sinks > 0 and 'position' in self.sinks.loaded_arrays():
            self._sink_arrays['position'] += translation

        if self.translation is None:
            self.translation = translation
        else:
            self.translation += translation

        return self

    def particle_indices(
        self, particle_type: str, squeeze: bool = False,
    ) -> Union[ndarray, List[ndarray]]:
        """Particle indices of a particular type.

        Parameters
        ----------
        particle_type
            The particle type as a string.
        squeeze
            If True return all subtypes in a single array. Default is
            False.

        Returns
        -------
        ndarray or list of ndarray
            If particle has no subtypes then returns an array of
            indices of that type. Otherwise return a list of arrays of
            indices, one for each subtype. However, if squeeze
            is True, return a single array.
        """
        if particle_type == 'dust' and not squeeze:
            # Dust particle sub-type skips zero: 1, 2, 3 ...
            return [
                np.flatnonzero(
                    (self['type'] == self.particle_type['dust'])
                    & (self['sub_type'] == idx + 1)
                )
                for idx in range(self.num_dust_species)
            ]
        if particle_type == 'boundary' and not squeeze:
            # Boundary particle sub-type: 0 (gas), 1, 2, 3... (dust)
            return [
                np.flatnonzero(
                    (self['type'] == self.particle_type['boundary'])
                    & (self['sub_type'] == idx)
                )
                for idx in range(self.num_dust_species + 1)
            ]
        return np.flatnonzero(self['type'] == self.particle_type[particle_type])

    def subsnaps_as_dict(
        self, squeeze: bool = False
    ) -> Dict[str, Union[SubSnap, List[SubSnap]]]:
        """Return particle-type subsnaps as a dict of SubSnaps.

        Parameters
        ----------
        squeeze : optional
            Squeeze sub-types. For each key, if False and the particle
            family has sub-types then return a list of SubSnaps of each
            sub-type. Otherwise return a SubSnap with all particle of
            that type.

        Returns
        -------
        Dict
            A dict of all SubSnaps.
        """
        subsnaps_as_dict: Dict[str, Any] = dict()
        for key in self.particle_type:
            try:
                subsnaps_as_dict[key] = self.family(key, squeeze=squeeze)
            except ValueError:
                continue

        return subsnaps_as_dict

    def subsnaps_as_list(self, squeeze: bool = False) -> List[SubSnap]:
        """Return particle-type subsnaps as a list of SubSnaps.

        Parameters
        ----------
        squeeze : optional
            Squeeze sub-types. If True then each particle sub-type of
            the same type will be treated the same.

        Returns
        -------
        List[SubSnap]
            A list of all SubSnaps.
        """
        subsnaps: List[SubSnap] = list()
        for val in self.subsnaps_as_dict(squeeze=squeeze).values():
            if isinstance(val, list):
                subsnaps = subsnaps + val
            else:
                subsnaps.append(val)
        return subsnaps

    def set_kernel(self, kernel: str) -> Snap:
        """Set kernel.

        Parameters
        ----------
        kernel
            The kernel name as a string.

        Returns
        -------
        Snap
            The Snap.
        """
        if kernel not in kernel_names:
            raise ValueError(f'Kernel must be in {kernel_names}')
        self._properties['kernel'] = kernel

        return self

    def set_gravitational_parameter(self, sink_idx: Union[int, List[int]]) -> Snap:
        """Set standard gravitational parameter.

        Calculate the standard gravitational parameter (G M) given a
        sink index, or list of sink indices for multiple systems, etc,
        and set snap.properties['gravitational_parameter']. This is
        required to calculate orbital quantities such as eccentricity
        or Stokes number.

        Parameters
        ----------
        sink_idx
            The sink index or list of indices.

        Returns
        -------
        Snap
            The Snap.
        """
        G = plonk_units.newtonian_constant_of_gravitation
        if isinstance(sink_idx, (int, list)):
            M = self.sinks[sink_idx]['mass']
        else:
            raise ValueError('Cannot determine gravitational parameter')
        G = (G * M).to_base_units()
        self._properties['gravitational_parameter'] = G

        return self

    def set_molecular_weight(self, molecular_weight: float) -> Snap:
        """Set molecular weight.

        Set the molecular weight of the gas in gram / mole in
        snap.properties['molecular_weight']. This is required to
        calculate temperature.

        Parameters
        ----------
        molecular_weight
            The molecular weight in units of gram / mole. E.g. Phantom
            uses 2.381 for molecular hydrogen with solar metallicity.

        Returns
        -------
        Snap
            The Snap.
        """
        self._properties['molecular_weight'] = molecular_weight
        return self

    @property
    def tree(self) -> cKDTree:
        """Particle neighbour kd-tree.

        Trees are represented by scipy cKDTree objects.
        """
        if self._tree is None:
            pos: Quantity = self['position']
            self._tree = cKDTree(pos.magnitude)
        return self._tree

    def neighbours(self, indices: Union[ndarray, List[int]]) -> ndarray:
        """Get neighbours of particles.

        Parameters
        ----------
        indices
            The particle indices.

        Returns
        -------
        ndarray of list
            An array of neighbours lists.
        """
        kernel = self._properties.get('kernel')
        if kernel is None:
            raise ValueError(
                'To calculate particle neighbours, first set the kernel\n'
                'via snap.set_kernel.'
            )

        subsnap = self[indices]
        position: Quantity = subsnap['position']
        smoothing_length: Quantity = subsnap['smoothing_length']
        r_kern = kernel_radius[kernel]

        neighbours = self.tree.query_ball_point(
            position.magnitude, r_kern * smoothing_length.magnitude, n_jobs=-1,
        )

        return neighbours

    def write_extra_arrays(
        self, arrays: List[str], filename: Union[str, Path] = None
    ) -> Snap:
        """Write extra arrays to file.

        Parameters
        ----------
        arrays
            A list of strings with array names.
        filename : optional
            A filename to write to.

        Returns
        -------
        Snap
            The Snap.
        """
        if filename is None:
            filename = f'{self.file_path.stem}_extra.h5'
        f = h5py.File(filename, mode='w')
        for array in arrays:
            arr_with_units: Quantity = self[array]
            units = self.array_code_unit(array)
            arr = (arr_with_units / units).magnitude
            dset = f.create_dataset(
                array, arr.shape, dtype=arr.dtype, compression='gzip',
            )
            dset[:] = arr
        f.close()

        return self

    def read_extra_arrays(self, filename: Union[str, Path] = None) -> Snap:
        """Read extra arrays from file.

        Parameters
        ----------
        filename : optional
            A filename to read from.

        Returns
        -------
        Snap
            The Snap.
        """
        if filename is None:
            filename = f'{self.file_path.stem}_extra.h5'
        f = h5py.File(filename, mode='r')
        for array in f:
            self[array] = f[array][()] * plonk_units('dimensionless')
        f.close()

        return self

    def to_dataframe(
        self, columns: Union[Tuple[str, ...], List[str]] = None, units: List[str] = None
    ) -> DataFrame:
        """Convert Snap to DataFrame.

        Parameters
        ----------
        columns : optional
            A list of columns to add to the data frame. Default is
            None.
        units : optional
            A list of units corresponding to columns add to the data
            frame. Units must be strings, and must be base units. I.e.
            'cm' not '10 cm'. If None, use default, i.e. cgs. Default is
            None.

        Returns
        -------
        DataFrame
        """
        d = dict()
        if columns is None:
            columns = self.loaded_arrays()
        cols = list(columns)
        if units is None:
            _units = list()
            for column in cols:
                try:
                    arr: Quantity = self[column]
                    _units.append(arr.units)
                except AttributeError:
                    _units.append(plonk_units('dimensionless'))
        else:
            _units = list()
            for unit in units:
                u = plonk_units(unit)
                if np.allclose(u.m, 1.0):
                    _units.append(u.units)
                else:
                    raise ValueError(
                        'Units must be strings, and must be base units. '
                        'I.e. "cm" not "10 cm".'
                    )
        if len(_units) != len(cols):
            raise ValueError('units and columns must have same length')
        for column, unit in zip(cols, _units):
            name = column
            array: Quantity = self[column]
            try:
                suffix = f' [{unit:~}]'
                array = array.to(unit).magnitude
            except AttributeError:
                suffix = ''
            if array.ndim == 1:
                d[name + suffix] = array
            if array.ndim == 2:
                for idx in range(array.shape[1]):
                    d[f'{name}.{idx+1}' + suffix] = array[:, idx]
        return pd.DataFrame(d)

    def family(self, name: str, squeeze: bool = False) -> Union[SubSnap, List[SubSnap]]:
        """Get a SubSnap of a particle family by name.

        Parameters
        ----------
        name
            A string representing the name of the family of particles,
            e.g. 'gas'.
        squeeze
            Squeeze sub-types. If False and the particle family has
            sub-types then return a list of SubSnaps of each sub-type.
            Otherwise return a SubSnap with all particle of that type.

        Returns
        -------
        SubSnap or List[SubSnap]
            If the particle family has no sub-types then return a
            SubSnap. Otherwise, if split_subtype is True, return a list
            of SubSnaps.
        """
        if name in self.particle_type:
            ind = self.particle_indices(particle_type=name, squeeze=squeeze)
            if isinstance(ind, list):
                if len(ind[0]) == 0:
                    raise ValueError('Family has no particles')
                return [SubSnap(self, _ind) for _ind in ind]
            if len(ind) == 0:
                raise ValueError('Family has no particles')
            return SubSnap(self, ind)
        raise ValueError('Family not available')

    def array(self, name: str, sinks: bool = False) -> Quantity:
        """Get a particle (or sink) array.

        Parameters
        ----------
        name
            A string representing the name of the particle array.
        sinks
            Whether or not to reference particle or sinks arrays.

        Returns
        -------
        Quantity
        """
        base_name = self.base_array_name(name)
        suffix = self._array_suffix(name)
        array = self._get_array(base_name, sinks)
        if suffix == '':
            return array
        transform, _slice, kwargs = self._array_transform(
            base_name=base_name, suffix=suffix
        )
        return transform(array, **kwargs)[_slice]

    def array_code_unit(self, arr: str) -> Quantity:
        """Get array code units.

        Parameters
        ----------
        arr
            The string representing the quantity.

        Returns
        -------
        Quantity
            The Pint unit quantity, or the float 1.0 if no unit found.
        """
        base_name = self.base_array_name(arr)
        try:
            unit = self._array_code_units[base_name]
        except KeyError:
            _arr: Quantity = self[base_name]
            dim = _arr.units.dimensionality
            unit = 1.0
            for d in self.code_units:
                unit *= self.code_units[d] ** dim[f'[{d}]']
        return unit

    def array_in_code_units(self, name: str) -> ndarray:
        """Get array in code units.

        Parameters
        ----------
        snap
            The Snap or SubSnap.
        name
            The array name.

        Returns
        -------
        ndarray
            The array on the particles in code units.
        """
        return (self[name] / self.array_code_unit(name)).to_reduced_units().magnitude

    def base_array_name(self, name: str) -> str:
        """Get the base array name from a string.

        For example, 'velocity_x' returns 'velocity', 'density' returns
        'density', 'dust_fraction_001' returns 'dust_fraction', 'x'
        returns 'position'.

        Parameters
        ----------
        name
            The name as a string

        Returns
        -------
        str
            The base array name.
        """
        if name in self.available_arrays():
            return name
        if self.num_sinks > 0 and name in self.sinks.available_arrays():
            return name
        if name in self._array_aliases:
            return self._array_aliases[name]
        if name in self._arrays:
            return name
        if name in self._sink_arrays:
            return name

        name_root = '_'.join(name.split('_')[:-1])
        name_suffix = name.split('_')[-1]
        if name_root == '' and name_suffix in ('x', 'y', 'z'):
            return 'position'
        if name_root in self._array_aliases:
            return self._array_aliases[name_root]
        if name_root in self._vector_arrays and name_suffix in ('x', 'y', 'z', 'mag'):
            return name_root
        if name_root in self._dust_arrays:
            if _str_is_int(name_suffix) or name_suffix == 'tot':
                return name_root

        raise ValueError('Unknown array')

    def _array_suffix(self, name: str) -> str:
        if name in self.available_arrays() or name in self._array_aliases:
            return ''
        if self.num_sinks > 0 and name in self.sinks.available_arrays():
            return ''
        if name in self._arrays:
            return ''
        if name in self._sink_arrays:
            return ''

        name_root = '_'.join(name.split('_')[:-1])
        name_suffix = name.split('_')[-1]
        if name_root == '' and name_suffix in ('x', 'y', 'z'):
            return name_suffix
        name_root = self.base_array_name(name_root)
        if name_root in self._vector_arrays:
            if name_suffix in ('x', 'y', 'z', 'mag'):
                return name_suffix
        if name_root in self._dust_arrays:
            if _str_is_int(name_suffix):
                return name_suffix
            if name_suffix == 'tot':
                return name_suffix

        raise ValueError('Unknown array')

    def _array_transform(
        self, base_name: str, suffix: str
    ) -> Tuple[Callable, tuple, Dict[str, Any]]:
        def nothing(x):
            return x

        if base_name in self._vector_arrays:
            if suffix == 'x':
                return nothing, (..., 0), {}
            if suffix == 'y':
                return nothing, (..., 1), {}
            if suffix == 'z':
                return nothing, (..., 2), {}
            if suffix == 'mag':
                return norm, (), {'axis': 1}
        if base_name in self._dust_arrays:
            if _str_is_int(suffix):
                return nothing, (..., int(suffix) - 1), {}
            if suffix == 'tot':
                return np.sum, (), {'axis': 1}

        raise ValueError('Unknown array')

    def _get_array_from_registry(self, name: str, sinks: bool = False) -> Quantity:
        if sinks:
            array = self._sink_registry[name](self)
        else:
            array = self._array_registry[name](self)
        if self.rotation is not None and name in self._vector_arrays:
            array_m, array_u = array.magnitude, array.units
            array = self.rotation.apply(array_m) * array_u
        if self.translation is not None and name == 'position':
            array += self.translation
        return array

    def _get_array(self, name: str, sinks: bool = False) -> Quantity:
        """Get an array by name."""
        if sinks:
            array_dict = self._sink_arrays
        else:
            array_dict = self._arrays
        if name in array_dict:
            if name in self.default_units:
                return array_dict[name].to(self.default_units[name])
            return array_dict[name]
        if name in self._array_registry or name in self._sink_registry:
            if name in self.default_units:
                array = self._get_array_from_registry(name, sinks).to(
                    self.default_units[name]
                )
            else:
                array = self._get_array_from_registry(name, sinks)
            if self.cache_arrays:
                if sinks:
                    self._sink_arrays[name] = array
                else:
                    self._arrays[name] = array
            return array
        raise ValueError('Array not available')

    def _getitem(
        self, inp: Union[str, ndarray, int, slice], sinks: bool = False,
    ) -> Union[Quantity, SubSnap, List[SubSnap]]:
        """Return an array, or family, or subset."""
        if isinstance(inp, str):
            if inp in self.particle_type:
                return self.family(name=inp)
            return self.array(name=inp, sinks=sinks)
        if sinks:
            raise ValueError('Cannot return sinks as SubSnap')
        return SubSnap(self, inp)

    def __getitem__(self, inp: Union[str, ndarray, int, slice]):
        """Return an array, or family, or subset."""
        return self._getitem(inp, sinks=False)

    def __setitem__(self, name: str, item: Quantity):
        """Set a particle array."""
        if not isinstance(item, Quantity):
            raise ValueError('"item" must be Pint Quantity')
        if item.shape[0] != len(self):
            raise ValueError('Length of array does not match particle number')
        if name in self.loaded_arrays():
            raise ValueError(
                'Attempting to overwrite existing array. To do so, first delete the '
                'array\nwith del snap["array"], then try again.'
            )
        if name in self._available_arrays() or name in self._array_aliases.keys():
            raise ValueError(
                'Attempting to set array already available. '
                'See snap.available_arrays().'
            )
        self._arrays[name] = item

    def __delitem__(self, name):
        """Delete an array from memory."""
        del self._arrays[name]

    def _ipython_key_completions_(self) -> List[str]:
        """Tab completion for IPython __getitem__ method."""
        return self.available_arrays(verbose=True)

    def __len__(self) -> int:
        """Length as number of particles."""
        return self.num_particles

    def __repr__(self) -> str:
        """Dunder repr method."""
        return self.__str__()

    def __str__(self) -> str:
        """Dunder str method."""
        return f'<plonk.Snap "{self.file_path.name}">'

    image = visualize.image
    plot = visualize.plot
    vector = visualize.vector
    context = context.context


class SubSnap(Snap):
    """A Snap subset of particles.

    A SubSnap can be generated from a Snap via an index array, a
    particle mask, or a string. SubSnaps can be used like a Snap,
    including: accessing arrays, plotting, finding neighbours, etc.

    Parameters
    ----------
    base
        The base Snap or SubSnap
    indices
        A (N,) array of particle indices to include in the SubSnap.

    Examples
    --------
    Generate a SubSnap directly.

    >>> subsnap = SubSnap(base=snap, indices=[0, 1, 2, 3])

    You can generate a SubSnap from a Snap object. For example, generate
    a SubSnap of the gas particles on a Snap.

    >>> subsnap = snap['gas']

    Generate a SubSnap of particles with a mask.

    >>> subsnap = snap[snap['x'] > 0]

    Generate a SubSnap of particles from indices.

    >>> subsnap = snap[:100]
    >>> subsnap = snap[[0, 9, 99]]
    """

    def __init__(self, base: Snap, indices: Union[ndarray, slice, list, int, tuple]):
        super().__init__()

        self.base = base

        ind = _input_indices_array(inp=indices, max_slice=len(base))
        if len(ind) == 0:
            logger.warning('SubSnap has no particles')
        self._indices = ind

        # Attributes different to Snap
        self._num_particles = len(self._indices)
        self._num_particles_of_type = -1
        self._num_dust_species = -1
        self._tree = None

        # Attributes same as Snap
        self.data_source = self.base.data_source
        self.file_path = self.base.file_path
        self._code_units = self.base._code_units
        self._default_units = self.base._default_units
        self._properties = self.base._properties
        self._array_code_units = self.base._array_code_units
        self._array_registry = self.base._array_registry
        self._sink_registry = self.base._sink_registry
        self._cache_arrays = self.base._cache_arrays
        self._arrays = self.base._arrays
        self._sink_arrays = self.base._sink_arrays
        self._file_pointer = self.base._file_pointer
        self.rotation = self.base.rotation
        self.translation = self.base.translation

    @property
    def indices(self) -> ndarray:
        """Particle indices."""
        return self._indices

    @property
    def sinks(self) -> Sinks:
        """Sink particle arrays."""
        return self.base.sinks

    def __repr__(self) -> str:
        """Dunder repr method."""
        return self.__str__()

    def __str__(self) -> str:
        """Dunder str method."""
        return f'<plonk.SubSnap "{self.file_path.name}">'

    def _get_array(self, name: str, sinks: bool = False) -> Quantity:
        return self.base._get_array(name, sinks)[self.indices]


SnapLike = Union[Snap, SubSnap]


class Sinks:
    """Sink particles in a Snap.

    A Sinks object is generated from a Snap.

    Parameters
    ----------
    base
        The base Snap.
    indices : optional
        Indices to specify a subset of sink particles.

    Examples
    --------
    Generate a Sinks object directly.

    >>> sinks = Sinks(base=snap)

    Generate a Sinks object from a Snap object.

    >>> sinks = snap.sinks

    Choose a subset of sink particles.

    >>> sinks = snap.sinks[[0, 1]]
    >>> star = snap.sinks[0]
    >>> planets = snap.sinks[1:4]
    """

    def __init__(
        self, base: Snap, indices: Union[ndarray, slice, list, int, tuple] = None
    ):
        self.base = base

        if indices is None:
            indices = np.arange(base.num_sinks)
        ind = _input_indices_array(inp=indices, max_slice=base.num_sinks)
        if len(ind) == 0:
            logger.warning('Sinks has no particles')
        self._indices = ind
        self._iter_index = 0

        # Attributes same as Snap
        self.file_path = self.base.file_path
        self.base_array_name = self.base.base_array_name
        self.default_units = self.base.default_units

    @property
    def indices(self) -> ndarray:
        """Sink particle indices."""
        return self._indices

    def available_arrays(
        self, verbose: bool = False, aliases: bool = False
    ) -> List[str]:
        """Return a list of available sink arrays.

        Parameters
        ----------
        verbose
            Also display suffixed arrays, e.g. 'position_x',
            'position_y', etc. Default is False
        aliases
            If True, return array aliases. Default is False.

        Returns
        -------
        List
            A list of names of available arrays.
        """
        return self.base._available_arrays(sinks=True, verbose=verbose, aliases=aliases)

    def loaded_arrays(self) -> List[str]:
        """Return a list of loaded arrays.

        Returns
        -------
        List
            A list of names of loaded sink arrays.
        """
        return sorted(self.base._sink_arrays.keys())

    def array(self, name: str) -> Quantity:
        """Get an array.

        Parameters
        ----------
        name
            A string representing the name of the particle array.

        Returns
        -------
        Quantity
        """
        return np.squeeze(self.base.array(name, sinks=True)[self.indices])[()]

    def __setitem__(self, name: str, item: Quantity):
        """Set an array."""
        if not isinstance(item, Quantity):
            raise ValueError('"item" must be an array with units, i.e. a Pint Quantity')
        if item.shape[0] != len(self):
            raise ValueError('Length of array does not match particle number')
        self.base._sink_arrays[name] = item

    def __getitem__(self, inp):
        """Return an array or subset."""
        if isinstance(inp, str):
            return self.array(name=inp)
        ind = _input_indices_array(inp=inp, max_slice=len(self))
        if ind is not None:
            _ind = self.indices[ind]
            return Sinks(self.base, _ind)
        raise ValueError('Cannot determine item to return')

    def __len__(self) -> int:
        """Length as number of particles."""
        return len(self.indices)

    def __repr__(self) -> str:
        """Dunder repr method."""
        return self.__str__()

    def __str__(self) -> str:
        """Dunder str method."""
        return f'<plonk.Sinks "{self.base.file_path.name}">'

    def __add__(self, other):
        """Add Sinks."""
        indices = list(self.indices) + list(other.indices)
        return self.base.sinks[indices]

    def __eq__(self, other):
        """Compare Sinks."""
        same_base = self.base == other.base
        same_indices = np.all(self.indices == other.indices)
        return same_base and same_indices

    def __iter__(self):
        """Sinks iterator."""
        return self

    def __next__(self):
        """Sinks iterator."""
        if self._iter_index > len(self) - 1:
            self._iter_index = 0
            raise StopIteration()
        index = self._iter_index
        self._iter_index += 1
        return self[index]

    def _ipython_key_completions_(self) -> List[str]:
        """Tab completion for IPython __getitem__ method."""
        return self.available_arrays(verbose=True)

    plot = visualize.plot


def _str_is_int(string: str) -> bool:
    try:
        int(string)
        return True
    except ValueError:
        return False


def _input_indices_array(
    inp: Union[ndarray, slice, list, int, tuple], max_slice: int
) -> Union[ndarray, List[int]]:
    """Take array, slice, int, list, tuple and return indices array."""
    if isinstance(inp, ndarray):
        if np.issubdtype(np.bool, inp.dtype):
            return np.flatnonzero(inp)
        if np.issubdtype(np.int, inp.dtype):
            return inp
    if isinstance(inp, (list, tuple)):
        return np.array(inp)
    if isinstance(inp, (int, np.integer)):
        return np.array([inp])
    if isinstance(inp, slice):
        i1 = inp.start if inp.start is not None else 0
        i2 = inp.stop if inp.stop is not None else max_slice
        return np.arange(i1, i2, inp.step)
    return []



"""Geometry, and coordinate transformations."""

from typing import Tuple

import numpy as np
from numpy import ndarray
from scipy.interpolate import RectBivariateSpline

try:
    from skimage import transform
except ImportError:
    transform = None

from .._logging import logger


def distance_from_plane(
    x: ndarray, y: ndarray, z: ndarray, normal: ndarray, height: float = 0
):
    """Calculate distance from a plane.

    Parameters
    ----------
    x
        The x-positions.
    y
        The y-positions.
    z
        The z-positions.
    normal
        The normal vector describing the plane (x, y, z).
    height
        The height of the plane above the origin.

    Return
    ------
    The distance from the plane of each point.
    """
    a, b, c = normal
    d = height
    return np.abs((a * x + b * y + c * z + d) / np.sqrt(a ** 2 + b ** 2 + c ** 2))


def cartesian_to_polar(
    interpolated_data_cartesian: ndarray,
    extent_cartesian: Tuple[float, float, float, float],
) -> Tuple[ndarray, Tuple[float, float, float, float]]:
    """Convert interpolated Cartesian pixel grid to polar coordinates.

    Parameters
    ----------
    interpolated_data_cartesian
        The interpolated data on a Cartesian grid.
    extent_cartesian
        The extent in Cartesian space as (xmin, xmax, ymin, ymax). It
        must be square.

    Returns
    -------
    interpolated_data_polar
        The interpolated data on a polar grid (R, phi).
    extent_polar
        The extent on a polar grid (0, Rmax, 0, 2π).
    """
    if transform is None:
        logger.error(
            'cartesian_to_polar requires skimage (scikit-image) which is unavailable\n'
            'try pip install scikit-image --or-- conda install scikit-image'
        )
    data, extent = interpolated_data_cartesian, extent_cartesian

    if not np.allclose(extent[1] - extent[0], extent[3] - extent[2]):
        raise ValueError('Bad polar plot: x and y have different scales')

    number_of_pixels = data.shape
    radius_pix = 0.5 * data.shape[0]

    data = transform.warp_polar(data, radius=radius_pix)

    radius = 0.5 * (extent[1] - extent[0])
    extent_polar = (0, radius, 0, 2 * np.pi)

    x_grid = np.linspace(*extent[:2], data.shape[0])
    y_grid = np.linspace(*extent[2:], data.shape[1])
    spl = RectBivariateSpline(x_grid, y_grid, data)
    x_regrid = np.linspace(extent[0], extent[1], number_of_pixels[0])
    y_regrid = np.linspace(extent[2], extent[3], number_of_pixels[1])
    interpolated_data_polar = spl(x_regrid, y_regrid)

    return interpolated_data_polar, extent_polar



"""Context manager for Snaps."""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .snap import Snap


class SnapCacheContext:
    """Context manager for caching particle arrays on a Snap.

    Parameters
    ----------
    snap
        The Snap object.
    cache
        Turn caching arrays on or off during a context.
    """

    def __init__(self, snap: Snap, cache: bool):
        self.snap = snap
        self.cache = cache
        self.previous_cache_arrays = snap.cache_arrays
        self.loaded_arrays = snap.loaded_arrays()

    def __enter__(self):
        """Enter context."""
        self.snap.cache_arrays = self.cache

    def __exit__(self, *args):
        """Exit context."""
        self.snap.cache_arrays = self.previous_cache_arrays
        if self.previous_cache_arrays is False:
            self.snap.bulk_unload(
                list(set(self.snap.loaded_arrays()).difference(self.loaded_arrays))
            )


def context(snap: Snap, cache: bool = True):
    """Context manager for caching particle arrays on a Snap.

    Caching arrays in memory improves performance by saving slow
    reads from disc. Caching arrays is turned on by default on a Snap.
    However, it can be useful to turn it off for low memory
    requirements, such as reading from many snapshots. In this case it
    may be useful to use this context manager to improve performance
    inside the context without then leaving those arrays in memory
    afterwards.

    Parameters
    ----------
    cache
        Turn caching arrays on or off during a context.

    Examples
    --------
    Temporarily turn on caching arrays.

    # Caching off
    >>> snap.cache_arrays
    False

    # No loaded arrays
    >>> snap.loaded_arrays()
    ()

    # Produce an image which loads arrays
    >>> with snap.context(cache=True):
    ...     ax = snap.image('density')

    # No loaded arrays
    >>> snap.loaded_arrays()
    ()
    """
    return SnapCacheContext(snap=snap, cache=cache)
